"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Generate synthetic tropical cyclone tracks from real ones
"""

import array
import itertools
import logging
import matplotlib.cm as cm_mp
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import numba
import numpy as np

from climada import CONFIG
import climada.util.coordinates
import climada.hazard.tc_tracks

LOGGER = logging.getLogger(__name__)

LANDFALL_DECAY_V = {
    -1: 0.00012859077693295416,
    0: 0.0017226346292718126,
    1: 0.002309772914350468,
    2: 0.0025968221565522698,
    3: 0.002626252944053856,
    4: 0.002550639312763181,
    5: 0.003788695795963695
}
"""Global landfall decay parameters for wind speed by TC category.
Keys are TC categories with -1='TD', 0='TS', 1='Cat 1', ..., 5='Cat 5'.
It is v_rel as derived from:
tracks = TCTracks.from_ibtracs_netcdf(year_range=(1980,2019), estimate_missing=True)
extent = tracks.get_extent()
land_geom = climada.util.coordinates.get_land_geometry(
    extent=extent, resolution=10
)
v_rel, p_rel = _calc_land_decay(tracks.data, land_geom, pool=tracks.pool)"""

LANDFALL_DECAY_P = {
    -1: (1.0088807492745373, 0.002117478217863062),
    0: (1.0192813768091684, 0.003068578025845065),
    1: (1.0362982218631644, 0.003620816186262243),
    2: (1.0468630800617038, 0.004067381088015585),
    3: (1.0639055205005432, 0.003708174876364079),
    4: (1.0828373148889825, 0.003997492773076179),
    5: (1.1088615145002092, 0.005224331234796362)}
"""Global landfall decay parameters for pressure by TC category.
Keys are TC categories with -1='TD', 0='TS', 1='Cat 1', ..., 5='Cat 5'.
It is p_rel as derived from:
tracks = TCTracks.from_ibtracs_netcdf(year_range=(1980,2019), estimate_missing=True)
extent = tracks.get_extent()
land_geom = climada.util.coordinates.get_land_geometry(
    extent=extent, resolution=10
)
v_rel, p_rel = _calc_land_decay(tracks.data, land_geom, pool=tracks.pool)"""

def calc_perturbed_trajectories(tracks,
                                nb_synth_tracks=9,
                                max_shift_ini=0.75,
                                max_dspeed_rel=0.3,
                                max_ddirection=np.pi / 360,
                                autocorr_dspeed=0.85,
                                autocorr_ddirection=0.5,
                                seed=CONFIG.hazard.trop_cyclone.random_seed.int(),
                                decay=True,
                                use_global_decay_params=True,
                                pool=None):
    """
    Generate synthetic tracks based on directed random walk. An ensemble of nb_synth_tracks
    synthetic tracks is computed for every track contained in self.

    The methodology perturbs the tracks locations, and if decay is True it additionally
    includes decay of wind speed and central pressure drop after landfall. No other track
    parameter is perturbed.
    The track starting point location is perturbed by random uniform values of
    magnitude up to max_shift_ini in both longitude and latitude. Then, each segment
    between two consecutive points is perturbed in direction and distance (i.e.,
    translational speed). These perturbations can be correlated in time, i.e.,
    the perturbation in direction applied to segment i is correlated with the perturbation
    in direction applied to segment i-1 (and similarly for the perturbation in translational
    speed).
    Perturbations in track direction and temporal auto-correlations in perturbations are
    on an hourly basis, and the perturbations in translational speed is relative.
    Hence, the parameter values are relatively insensitive to the temporal
    resolution of the tracks. Note however that all tracks should be at the same
    temporal resolution, which can be achieved using equal_timestep().
    max_dspeed_rel and autocorr_dspeed control the spread along the track ('what distance
    does the track run for'), while max_ddirection and autocorr_ddirection control the spread
    perpendicular to the track movement ('how does the track diverge in direction').
    max_dspeed_rel and max_ddirection control the amplitude of perturbations at each track
    timestep but perturbations may tend to compensate each other over time, leading to
    a similar location at the end of the track, while autocorr_dspeed and autocorr_ddirection
    control how these perturbations persist in time and hence the amplitude of the
    perturbations towards the end of the track.

    Note that the default parameter values have been only roughly calibrated so that
    the frequency of tracks in each 5x5degree box remains approximately constant.
    This is not an in-depth calibration and should be treated as such.
    The object is mutated in-place.

    Parameters
    ----------
    tracks : climada.hazard.TCTracks
        Tracks data.
    nb_synth_tracks : int, optional
        Number of ensemble members per track. Default: 9.
    max_shift_ini : float, optional
        Amplitude of max random starting point shift in decimal degree
        (up to +/-max_shift_ini for longitude and latitude). Default: 0.75.
    max_dspeed_rel : float, optional
        Amplitude of translation speed perturbation in relative terms
        (e.g., 0.2 for +/-20%). Default: 0.3.
    max_ddirection : float, optional
        Amplitude of track direction (bearing angle) perturbation
        per hour, in radians. Default: pi/360.
    autocorr_dspeed : float, optional
        Temporal autocorrelation in translation speed perturbation
        at a lag of 1 hour. Default: 0.85.
    autocorr_ddirection : float, optional
        Temporal autocorrelation of translational direction perturbation
        at a lag of 1 hour. Default: 0.5.
    seed : int, optional
        Random number generator seed for replicability of random walk.
        Put negative value if you don't want to use it. Default: configuration file.
    decay : bool, optional
        Whether to apply landfall decay in probabilistic tracks. Default: True.
    use_global_decay_params : bool, optional
        Whether to use precomputed global parameter values for landfall decay
        obtained from IBTrACS (1980-2019). If False, parameters are fitted
        using historical tracks in input parameter 'tracks', in which case the
        landfall decay applied depends on the tracks passed as an input and may
        not be robust if few historical tracks make landfall in this object.
        Default: True.
    pool : pathos.pool, optional
        Pool that will be used for parallel computation when applicable. If not given, the
        pool attribute of `tracks` will be used. Default: None
    """
    LOGGER.info('Computing %s synthetic tracks.', nb_synth_tracks * tracks.size)

    pool = tracks.pool if pool is None else pool

    if seed >= 0:
        np.random.seed(seed)

    # ensure tracks have constant time steps
    time_step_h = np.unique(np.concatenate([np.unique(x['time_step']) for x in tracks.data]))
    if not np.allclose(time_step_h, time_step_h[0]):
        raise ValueError('Tracks have different temporal resolution. '
                         'Please ensure constant time steps by applying equal_timestep beforehand')
    time_step_h = time_step_h[0]

    # number of random value per synthetic track:
    # 2*nb_synth_tracks for starting points (lon, lat)
    # nb_synth_tracks*(track.time.size-1) for angle and same for translation perturbation
    # hence sum is nb_synth_tracks * (2 + 2*(size-1)) = nb_synth_tracks * 2 * size
    # https://stats.stackexchange.com/questions/48086/algorithm-to-produce-autocorrelated-uniformly-distributed-number
    if autocorr_ddirection == 0 and autocorr_dspeed == 0:
        random_vec = [np.random.uniform(size=nb_synth_tracks * (2 * track.time.size))
                      for track in tracks.data]
    else:
        random_vec = [np.concatenate((np.random.uniform(size=nb_synth_tracks * 2),
                                      _random_uniform_ac(nb_synth_tracks * (track.time.size - 1),
                                                         autocorr_ddirection, time_step_h),
                                      _random_uniform_ac(nb_synth_tracks * (track.time.size - 1),
                                                         autocorr_dspeed, time_step_h)))
                      if track.time.size > 1 else np.random.uniform(size=nb_synth_tracks * 2)
                      for track in tracks.data]

    if pool:
        chunksize = min(tracks.size // pool.ncpus, 1000)
        new_ens = pool.map(_one_rnd_walk, tracks.data,
                           itertools.repeat(nb_synth_tracks, tracks.size),
                           itertools.repeat(max_shift_ini, tracks.size),
                           itertools.repeat(max_dspeed_rel, tracks.size),
                           itertools.repeat(max_ddirection, tracks.size),
                           random_vec, chunksize=chunksize)
    else:
        new_ens = [_one_rnd_walk(track, nb_synth_tracks, max_shift_ini,
                                 max_dspeed_rel, max_ddirection, rand)
                   for track, rand in zip(tracks.data, random_vec)]

    cutoff_track_ids_tc = [x[1] for x in new_ens]
    cutoff_track_ids_tc = sum(cutoff_track_ids_tc, [])
    cutoff_track_ids_ts = [x[2] for x in new_ens]
    cutoff_track_ids_ts = sum(cutoff_track_ids_ts, [])
    if len(cutoff_track_ids_tc) > 0:
        LOGGER.info('The following generated synthetic tracks moved beyond '
                    'the range of [-70, 70] degrees latitude. Cut out '
                    'at TC category >1: %s.',
                    ', '.join(cutoff_track_ids_tc))
    if len(cutoff_track_ids_ts) > 0:
        LOGGER.debug('The following generated synthetic tracks moved beyond '
                     'the range of [-70, 70] degrees latitude. Cut out '
                     'at TC category <= 1: %s.',
                     ', '.join(cutoff_track_ids_ts))
    new_ens = [x[0] for x in new_ens]
    tracks.data = sum(new_ens, [])

    if decay:
        extent = tracks.get_extent()
        land_geom = climada.util.coordinates.get_land_geometry(
            extent=extent, resolution=10
        )
        if use_global_decay_params:
            tracks.data = _apply_land_decay(tracks.data, LANDFALL_DECAY_V,
                                            LANDFALL_DECAY_P, land_geom, pool=pool)
        else:
            # fit land decay coefficients based on historical tracks
            hist_tracks = [track for track in tracks.data if track.orig_event_flag]
            if hist_tracks:
                try:
                    v_rel, p_rel = _calc_land_decay(hist_tracks, land_geom, pool=pool)
                    tracks.data = _apply_land_decay(
                        tracks.data, v_rel, p_rel, land_geom, pool=pool)
                except ValueError as verr:
                    raise ValueError('Landfall decay could not be applied.') from verr
            else:
               raise ValueError('No historical tracks found. Historical'
                                'tracks are needed for land decay calibration'
                                'if use_global_decay_params=False.')


def _one_rnd_walk(track, nb_synth_tracks, max_shift_ini, max_dspeed_rel, max_ddirection, rnd_vec):
    """
    Apply random walk to one track.

    Parameters
    ----------
    track : xr.Dataset
        Track data.
    nb_synth_tracks : int, optional
        Number of ensemble members per track. Default: 9.
    max_shift_ini : float, optional
        Amplitude of max random starting point shift in decimal degree
        (up to +/-max_shift_ini for longitude and latitude). Default: 0.75.
    max_dspeed_rel : float, optional
        Amplitude of translation speed perturbation in relative terms
        (e.g., 0.2 for +/-20%). Default: 0.3.
    max_ddirection : float, optional
        Amplitude of track direction (bearing angle) perturbation
        per hour, in radians. Default: pi/180.
    rnd_vec : np.ndarray of shape (2 * nb_synth_tracks * track.time.size),)
        Vector of random perturbations.

    Returns
    -------
    ens_track : list(xr.Dataset)
        List of the track and the generated synthetic tracks.
    cutoff_track_ids_tc : List of str
        List containing information about the tracks that were cut off at high
        latitudes with wind speed of TC category 2-5.
    curoff_track_ids_ts : List of str
        List containing information about the tracks that were cut off at high
        latitudes with a wind speed up to TC category 1.
    """
    ens_track = list()
    n_dat = track.time.size
    n_seg = n_dat - 1
    xy_ini = max_shift_ini * (2 * rnd_vec[:2 * nb_synth_tracks].reshape((2, nb_synth_tracks)) - 1)
    [dt] = np.unique(track['time_step'])

    ens_track.append(track)
    cutoff_track_ids_ts = []
    cutoff_track_ids_tc = []
    for i_ens in range(nb_synth_tracks):
        i_track = track.copy(True)

        # select angular perturbation for that synthetic track
        i_start_ang = 2 * nb_synth_tracks + i_ens * n_seg
        i_end_ang = i_start_ang + track.time.size - 1
        # scale by maximum perturbation and time step in hour (temporal-resolution independent)
        ang_pert = dt * np.degrees(max_ddirection * (2 * rnd_vec[i_start_ang:i_end_ang] - 1))
        ang_pert_cum = np.cumsum(ang_pert)

        # select translational speed perturbation for that synthetic track
        i_start_trans = 2 * nb_synth_tracks + nb_synth_tracks * n_seg + i_ens * n_seg
        i_end_trans = i_start_trans + track.time.size - 1
        # scale by maximum perturbation and time step in hour (temporal-resolution independent)
        trans_pert = 1 + max_dspeed_rel * (2 * rnd_vec[i_start_trans:i_end_trans] - 1)

        # get bearings and angular distance for the original track
        bearings = _get_bearing_angle(i_track.lon.values, i_track.lat.values)
        angular_dist = climada.util.coordinates.dist_approx(i_track.lat.values[:-1, None],
                                                            i_track.lon.values[:-1, None],
                                                            i_track.lat.values[1:, None],
                                                            i_track.lon.values[1:, None],
                                                            method="geosphere",
                                                            units="degree")[:, 0, 0]

        # apply perturbation to lon / lat
        new_lon = np.zeros_like(i_track.lon.values)
        new_lat = np.zeros_like(i_track.lat.values)
        new_lon[0] = i_track.lon.values[0] + xy_ini[0, i_ens]
        new_lat[0] = i_track.lat.values[0] + xy_ini[1, i_ens]
        last_idx = i_track.time.size
        for i in range(0, len(new_lon) - 1):
            new_lon[i + 1], new_lat[i + 1] = \
                _get_destination_points(new_lon[i], new_lat[i],
                                        bearings[i] + ang_pert_cum[i],
                                        trans_pert[i] * angular_dist[i])
            # if track crosses latitudinal thresholds (+-70°),
            # keep up to this segment (i+1), set i+2 as last point,
            # and discard all further points > i+2.
            if i+2 < last_idx and (new_lat[i + 1] > 70 or new_lat[i + 1] < -70):
                last_idx = i + 2
                # end the track here
                max_wind_end = i_track.max_sustained_wind.values[last_idx]
                ss_scale_end = climada.hazard.tc_tracks.set_category(max_wind_end,
                                                                     i_track.max_sustained_wind_unit)
                # TC category at ending point should not be higher than 1
                cutoff_txt = i_track.attrs['name'] + '_gen' + str(i_ens + 1)
                cutoff_txt = cutoff_txt + ' (%s)' % climada.hazard.tc_tracks.CAT_NAMES[ss_scale_end]
                if ss_scale_end > 1:
                    cutoff_track_ids_tc = cutoff_track_ids_tc + [cutoff_txt]
                else:
                    cutoff_track_ids_ts = cutoff_track_ids_ts + [cutoff_txt]
                break
        # make sure longitude values are within (-180, 180)
        climada.util.coordinates.lon_normalize(new_lon, center=0.0)

        i_track.lon.values = new_lon
        i_track.lat.values = new_lat
        i_track.attrs['orig_event_flag'] = False
        i_track.attrs['name'] = i_track.attrs['name'] + '_gen' + str(i_ens + 1)
        i_track.attrs['sid'] = i_track.attrs['sid'] + '_gen' + str(i_ens + 1)
        i_track.attrs['id_no'] = i_track.attrs['id_no'] + (i_ens + 1) / 100
        i_track = i_track.isel(time=slice(None, last_idx))

        ens_track.append(i_track)

    return ens_track, cutoff_track_ids_tc, cutoff_track_ids_ts


def _random_uniform_ac(n_ts, autocorr, time_step_h):
    """
    Generate a series of autocorrelated uniformly distributed random numbers.

    This implements the algorithm described here to derive a uniformly distributed
    series with specified autocorrelation (here at a lag of 1 hour):
    https://stats.stackexchange.com/questions/48086/
        algorithm-to-produce-autocorrelated-uniformly-distributed-number
    Autocorrelation is specified at a lag of 1 hour. To get a time series at a
    different temporal resolution (time_step_h), an hourly time series is generated
    and resampled (using linear interpolation) to the target resolution.

    Parameters
    ----------
    n_ts : int
        Length of the series.
    autocorr : float
        Autocorrelation (between -1 and 1) at hourly time scale.
    time_step_h : float
        Temporal resolution of the time series, in hour.

    Returns
    -------
    x_ts : numpy.ndarray of shape (n_ts,)
        n values at time_step_h intervals that are uniformly distributed and with
            the requested temporal autocorrelation at a scale of 1 hour.
    """
    # generate autocorrelated 1-hourly perturbations, so first create hourly
    #   time series of perturbations
    n_ts_hourly_exact = n_ts * time_step_h
    n_ts_hourly = int(np.ceil(n_ts_hourly_exact))
    x = np.random.normal(size=n_ts_hourly)
    theta = np.arccos(autocorr)
    for i in range(1, len(x)):
        x[i] = _h_ac(x[i - 1], x[i], theta)
    # scale x to have magnitude [0,1]
    x = (x + np.sqrt(3)) / (2 * np.sqrt(3))
    # resample at target time step
    x_ts = np.interp(np.arange(start=0, stop=n_ts_hourly_exact, step=time_step_h),
                     np.arange(n_ts_hourly), x)
    return x_ts


@numba.njit
def _h_ac(x, y, theta):
    """
    Generate next random number from current number for autocorrelated uniform series

    Implements function h defined in:
    https://stats.stackexchange.com/questions/48086/
        algorithm-to-produce-autocorrelated-uniformly-distributed-number

    Parameters
    ----------
    x : float
        Previous random number.
    y : float
        Random Standard Normal.
    theta : float
        arccos of autocorrelation.

    Returns
    -------
    x_next : float
        Next value in the series.
    """
    gamma = np.abs(np.mod(theta, np.pi) - \
                   np.floor((np.mod(theta, np.pi) / (np.pi / 2)) + 0.5) * np.pi / 2)
    x_next = 2 * np.sqrt(3) * (_f_ac(np.cos(theta) * x + np.sin(theta) * y, gamma) - 1 / 2)
    return x_next


@numba.njit
def _f_ac(z, theta):
    """
    F transform for autocorrelated random uniform series generation

    Implements function F defined in:
    https://stats.stackexchange.com/questions/48086/
        algorithm-to-produce-autocorrelated-uniformly-distributed-number
    i.e., the CDF of Y.

    Parameters
    ----------
    z : float
        Value.
    theta : float
        arccos of autocorrelation.

    Returns
    -------
        res : float
            CDF at value z
    """
    c = np.cos(theta)
    s = np.sin(theta)
    if z >= np.sqrt(3) * (c + s):
        res = 1
    elif z > np.sqrt(3) * (c - s):
        res = 1 / 12 / np.sin(2 * theta) * \
              (-3 - z ** 2 + 2 * np.sqrt(3) * z * (c + s) + 9 * np.sin(2 * theta))
    elif z > np.sqrt(3) * (-c + s):
        res = 1 / 6 * (3 + np.sqrt(3) * z / c)
    elif z > -np.sqrt(3) * (c + s):
        res = 1 / 12 / np.sin(2 * theta) * \
              (z ** 2 + 2 * np.sqrt(3) * z * (c + s) + 3 * (1 + np.sin(2 * theta)))
    else:
        res = 0
    return res


@numba.njit
def _get_bearing_angle(lon, lat):
    """
    Compute bearing angle of great circle paths defined by consecutive points

    Returns initial bearing (also called forward azimuth) of the n-1 great circle
    paths define by n consecutive longitude/latitude points. The bearing is the angle
    (clockwise from North) which if followed in a straight line along a great-circle
    arc will take you from the start point to the end point. See also:
    http://www.movable-type.co.uk/scripts/latlong.html
    Here, the bearing of each pair of consecutive points is computed.

    Parameters
    ----------
    lon : numpy.ndarray of shape (n,)
        Longitude coordinates of consecutive point, in decimal degrees.
    lat : numpy.ndarray of shape (n,)
        Latitude coordinates of consecutive point, in decimal degrees.

    Returns
    -------
        earth_ang_fix : numpy.ndarray of shape (n-1,)
            Bearing angle for each segment, in decimal degrees
    """
    lon, lat = map(np.radians, [lon, lat])
    # Segments between all point (0 -> 1, 1 -> 2, ..., n-1 -> n)
    # starting points
    lat_1 = lat[:-1]
    lon_1 = lon[:-1]
    # ending points
    lat_2 = lat[1:]
    lon_2 = lon[1:]
    delta_lon = lon_2 - lon_1
    # what to do with the points that don't move?
    #   i.e. where lat_2=lat_1 and lon_2=lon_1? The angle does not matter in
    # that case because angular distance will be 0.
    earth_ang_fix = np.arctan2(np.sin(delta_lon) * np.cos(lat_2),
                               np.cos(lat_1) * np.sin(lat_2) - \
                               np.sin(lat_1) * np.cos(lat_2) * np.cos(delta_lon))
    return np.degrees(earth_ang_fix)


@numba.njit
def _get_destination_points(lon, lat, bearing, angular_distance):
    """
    Get coordinates of endpoints from a given locations with the provided bearing and distance

    Parameters
    ----------
    lon : numpy.ndarray of shape (n,)
        Longitude coordinates of each starting point, in decimal degrees.
    lat : numpy.ndarray of shape (n,)
        Latitude coordinates of each starting point, in decimal degrees.
    bearing : numpy.ndarray of shape (n,)
        Bearing to follow for each starting point (direction Northward, clockwise).
    angular_distance : numpy.ndarray of shape (n,)
        Angular distance to travel for each starting point, in decimal degrees.

    Returns
    -------
        lon_2 : numpy.ndarray of shape (n,)
            Longitude coordinates of each ending point, in decimal degrees.
        lat_2 : numpy.ndarray of shape (n,)
            Latitude coordinates of each ending point, in decimal degrees.
    """
    lon, lat = map(np.radians, [lon, lat])
    bearing = np.radians(bearing)
    angular_distance = np.radians(angular_distance)
    lat_2 = np.arcsin(np.sin(lat) * np.cos(angular_distance) + np.cos(lat) * \
                      np.sin(angular_distance) * np.cos(bearing))
    lon_2 = lon + np.arctan2(np.sin(bearing) * np.sin(angular_distance) * np.cos(lat),
                             np.cos(angular_distance) - np.sin(lat) * np.sin(lat_2))
    return np.degrees(lon_2), np.degrees(lat_2)


def _calc_land_decay(hist_tracks, land_geom, s_rel=True, check_plot=False,
                     pool=None):
    """Compute wind and pressure decay coefficients from historical events

    Decay is calculated for every TC category according to the formulas:

        - wind decay = exp(-x*A)
        - pressure decay = S-(S-1)*exp(-x*B)

    Parameters
    ----------
    hist_tracks : list
        List of xarray Datasets describing TC tracks.
    land_geom : shapely.geometry.multipolygon.MultiPolygon
        land geometry
    s_rel : bool, optional
        use environmental presure to calc S value
        (true) or central presure (false)
    check_plot : bool, optional
        visualize computed coefficients.
        Default: False

    Returns
    -------
    v_rel : dict(category: A)
    p_rel : dict(category: (S, B))
    """

    if len(hist_tracks) < 100:
        LOGGER.warning('For the calibration of the landfall decay '
                       'it is recommended to provide as many historical '
                       'tracks as possible, but only %s historical tracks '
                       'were provided. '
                       'For a more robust calculation consider using '
                       'a larger number of tracks or set '
                       '`use_global_decay_params` to True', len(hist_tracks))

    # Key is Saffir-Simpson scale
    # values are lists of wind/wind at landfall
    v_lf = dict()
    # values are tuples with first value the S parameter, second value
    # list of central pressure/central pressure at landfall
    p_lf = dict()
    # x-scale values to compute landfall decay
    x_val = dict()

    if pool:
        dec_val = pool.map(_decay_values, hist_tracks, itertools.repeat(land_geom),
                           itertools.repeat(s_rel),
                           chunksize=min(len(hist_tracks) // pool.ncpus, 1000))
    else:
        dec_val = [_decay_values(track, land_geom, s_rel) for track in hist_tracks]

    for (tv_lf, tp_lf, tx_val) in dec_val:
        for key in tv_lf.keys():
            v_lf.setdefault(key, []).extend(tv_lf[key])
            p_lf.setdefault(key, ([], []))
            p_lf[key][0].extend(tp_lf[key][0])
            p_lf[key][1].extend(tp_lf[key][1])
            x_val.setdefault(key, []).extend(tx_val[key])

    v_rel, p_rel = _decay_calc_coeff(x_val, v_lf, p_lf)
    if check_plot:
        _check_decay_values_plot(x_val, v_lf, p_lf, v_rel, p_rel)

    return v_rel, p_rel


def _apply_land_decay(tracks, v_rel, p_rel, land_geom, s_rel=True,
                      check_plot=False, pool=None):
    """Compute wind and pressure decay due to landfall in synthetic tracks.

    Parameters
    ----------
    v_rel : dict
        {category: A}, where wind decay = exp(-x*A)
    p_rel : dict
        (category: (S, B)}, where pressure decay
        = S-(S-1)*exp(-x*B)
    land_geom : shapely.geometry.multipolygon.MultiPolygon
        land geometry
    s_rel : bool, optional
        use environmental presure to calc S value
        (true) or central presure (false)
    check_plot : bool, optional
        visualize computed changes
    """
    sy_tracks = [track for track in tracks if not track.orig_event_flag]
    if not sy_tracks:
        raise ValueError('No synthetic tracks contained. Synthetic tracks'
                         ' are needed.')

    if not v_rel or not p_rel:
        LOGGER.info('No decay coefficients.')
        return

    if check_plot:
        orig_wind, orig_pres = [], []
        for track in sy_tracks:
            orig_wind.append(np.copy(track.max_sustained_wind.values))
            orig_pres.append(np.copy(track.central_pressure.values))

    if pool:
        chunksize = min(len(tracks) // pool.ncpus, 1000)
        tracks = pool.map(_apply_decay_coeffs, tracks,
                          itertools.repeat(v_rel), itertools.repeat(p_rel),
                          itertools.repeat(land_geom), itertools.repeat(s_rel),
                          chunksize=chunksize)
    else:
        tracks = [_apply_decay_coeffs(track, v_rel, p_rel, land_geom, s_rel)
                  for track in tracks]

    for track in tracks:
        if track.orig_event_flag:
            climada.hazard.tc_tracks.track_land_params(track, land_geom)
    if check_plot:
        _check_apply_decay_plot(tracks, orig_wind, orig_pres)
    return tracks


def _decay_values(track, land_geom, s_rel):
    """Compute wind and pressure relative to landafall values.

    Parameters
    ----------
    track : xr.Dataset
        track
    land_geom : shapely.geometry.multipolygon.MultiPolygon
        land geometry
    s_rel : bool
        use environmental presure for S value (true) or central presure (false)

    Returns
    -------
    v_lf : dict
        key is Saffir-Simpson scale, values are arrays of wind/wind at landfall
    p_lf : dict
        key is Saffir-Simpson scale, values are tuples with first value array
        of S parameter, second value array of central pressure/central pressure
        at landfall
    x_val : dict
        key is Saffir-Simpson scale, values are arrays with the values used as
        "x" in the coefficient fitting, the distance since landfall
    """
    v_lf = dict()
    p_lf = dict()
    x_val = dict()

    climada.hazard.tc_tracks.track_land_params(track, land_geom)
    sea_land_idx, land_sea_idx = climada.hazard.tc_tracks._get_landfall_idx(track)
    if sea_land_idx.size:
        for sea_land, land_sea in zip(sea_land_idx, land_sea_idx):
            v_landfall = track.max_sustained_wind[sea_land - 1].values
            ss_scale = climada.hazard.tc_tracks.set_category(v_landfall,
                                                             track.max_sustained_wind_unit)

            v_land = track.max_sustained_wind[sea_land - 1:land_sea].values
            if v_land[0] > 0:
                v_land = (v_land[1:] / v_land[0]).tolist()
            else:
                v_land = v_land[1:].tolist()

            p_landfall = float(track.central_pressure[sea_land - 1].values)
            p_land = track.central_pressure[sea_land - 1:land_sea].values
            p_land = (p_land[1:] / p_land[0]).tolist()

            p_land_s = _calc_decay_ps_value(
                track, p_landfall, land_sea - 1, s_rel)
            p_land_s = len(p_land) * [p_land_s]

            if ss_scale not in v_lf:
                v_lf[ss_scale] = array.array('f', v_land)
                p_lf[ss_scale] = (array.array('f', p_land_s),
                                      array.array('f', p_land))
                x_val[ss_scale] = array.array('f',
                                                  track.dist_since_lf[sea_land:land_sea])
            else:
                v_lf[ss_scale].extend(v_land)
                p_lf[ss_scale][0].extend(p_land_s)
                p_lf[ss_scale][1].extend(p_land)
                x_val[ss_scale].extend(track.dist_since_lf[sea_land:land_sea])
    return v_lf, p_lf, x_val


def _decay_calc_coeff(x_val, v_lf, p_lf):
    """From track's relative velocity and pressure, compute the decay
    coefficients.
    - wind decay = exp(-x*A)
    - pressure decay = S-(S-1)*exp(-x*A)

    Parameters
    ----------
    x_val : dict
        key is Saffir-Simpson scale, values are lists with
        the values used as "x" in the coefficient fitting, the
        distance since landfall
    v_lf : dict
        key is Saffir-Simpson scale, values are lists of
        wind/wind at landfall
    p_lf : dict
        key is Saffir-Simpson scale, values are tuples with
        first value the S parameter, second value list of central
        pressure/central pressure at landfall

    Returns
    -------
    v_rel : dict
    p_rel : dict
    """
    np.warnings.filterwarnings('ignore')
    v_rel = dict()
    p_rel = dict()
    for ss_scale, val_lf in v_lf.items():
        x_val_ss = np.array(x_val[ss_scale])

        y_val = np.array(val_lf)
        v_coef = _solve_decay_v_function(y_val, x_val_ss)
        v_coef = v_coef[np.isfinite(v_coef)]
        v_coef = np.mean(v_coef)

        ps_y_val = np.array(p_lf[ss_scale][0])
        y_val = np.array(p_lf[ss_scale][1])
        y_val[ps_y_val <= y_val] = np.nan
        y_val[ps_y_val <= 1] = np.nan
        valid_p = np.isfinite(y_val)
        ps_y_val = ps_y_val[valid_p]
        y_val = y_val[valid_p]
        p_coef = _solve_decay_p_function(ps_y_val, y_val, x_val_ss[valid_p])
        ps_y_val = np.mean(ps_y_val)
        p_coef = np.mean(p_coef)

        if np.isfinite(v_coef) and np.isfinite(ps_y_val) and np.isfinite(ps_y_val):
            v_rel[ss_scale] = v_coef
            p_rel[ss_scale] = (ps_y_val, p_coef)

    scale_fill = np.array(list(p_rel.keys()))
    if not scale_fill.size:
        LOGGER.info('No historical track with landfall.')
        return v_rel, p_rel
    for ss_scale in climada.hazard.tc_tracks.CAT_NAMES.keys():
        if ss_scale not in p_rel:
            close_scale = scale_fill[np.argmin(np.abs(scale_fill - ss_scale))]
            LOGGER.debug('No historical track of category %s with landfall. '
                         'Decay parameters from category %s taken.',
                         climada.hazard.tc_tracks.CAT_NAMES[ss_scale],
                         climada.hazard.tc_tracks.CAT_NAMES[close_scale])
            v_rel[ss_scale] = v_rel[close_scale]
            p_rel[ss_scale] = p_rel[close_scale]
        elif v_rel[ss_scale] < 0:
            raise ValueError('The calibration of landfall decay for wind '
                             'speed resulted in a wind speed increase for '
                             'TC category %s.'
                             'This behavious in unphysical.'
                             'Please use a larger number of tracks or '
                             'use global paramaters by setting '
                             '`use_global_decay_params` to True',
                             climada.hazard.tc_tracks.CAT_NAMES[ss_scale])
        elif p_rel[ss_scale][0] < 0 or p_rel[ss_scale][1] < 0:
            raise ValueError('The calibration of landfall decay for central '
                             'pressure resulted in a pressure decrease for '
                             'TC category %s.'
                             'This behavious in unphysical.'
                             'Please use a larger number of tracks or '
                             'use global paramaters by setting '
                             '`use_global_decay_params` to True',
                             climada.hazard.tc_tracks.CAT_NAMES[ss_scale])

    return v_rel, p_rel


def _check_decay_values_plot(x_val, v_lf, p_lf, v_rel, p_rel):
    """Generate one graph with wind decay and an other with central pressure
    decay, true and approximated."""
    # One graph per TC category
    for track_cat, color in zip(v_lf.keys(),
                                cm_mp.rainbow(np.linspace(0, 1, len(v_lf)))):
        _, axes = plt.subplots(2, 1)
        x_eval = np.linspace(0, np.max(x_val[track_cat]), 20)

        axes[0].set_xlabel('Distance from landfall (km)')
        axes[0].set_ylabel('Max sustained wind\nrelative to landfall')
        axes[0].set_title('Wind, TC cat %s' %
                          climada.hazard.tc_tracks.CAT_NAMES[track_cat])
        axes[0].plot(x_val[track_cat], v_lf[track_cat], '*', c=color,
                     label=climada.hazard.tc_tracks.CAT_NAMES[track_cat])
        axes[0].plot(x_eval, _decay_v_function(v_rel[track_cat], x_eval),
                     '-', c=color)

        axes[1].set_xlabel('Distance from landfall (km)')
        axes[1].set_ylabel('Central pressure\nrelative to landfall')
        axes[1].set_title('Pressure, TC cat %s' %
                          climada.hazard.tc_tracks.CAT_NAMES[track_cat])
        axes[1].plot(x_val[track_cat], p_lf[track_cat][1], '*', c=color,
                     label=climada.hazard.tc_tracks.CAT_NAMES[track_cat])
        axes[1].plot(
            x_eval,
            _decay_p_function(p_rel[track_cat][0], p_rel[track_cat][1], x_eval),
            '-', c=color)


def _apply_decay_coeffs(track, v_rel, p_rel, land_geom, s_rel):
    """Change track's max sustained wind and central pressure using the land
    decay coefficients.

    Parameters
    ----------
    track : xr.Dataset
        TC track
    v_rel : dict
        {category: A}, where wind decay = exp(-x*A)
    p_rel : dict
        (category: (S, B)},
        where pressure decay = S-(S-1)*exp(-x*B)
    land_geom : shapely.geometry.multipolygon.MultiPolygon
        land geometry
    s_rel : bool
        use environmental presure for S value (true) or
        central presure (false)

    Returns
    -------
    xr.Dataset
    """
    # return if historical track
    if track.orig_event_flag:
        return track

    climada.hazard.tc_tracks.track_land_params(track, land_geom)
    sea_land_idx, land_sea_idx = climada.hazard.tc_tracks._get_landfall_idx(track)
    if not sea_land_idx.size:
        return track
    for idx, (sea_land, land_sea) \
            in enumerate(zip(sea_land_idx, land_sea_idx)):
        v_landfall = track.max_sustained_wind[sea_land - 1].values
        p_landfall = float(track.central_pressure[sea_land - 1].values)
        ss_scale = climada.hazard.tc_tracks.set_category(v_landfall,
                                                         track.max_sustained_wind_unit)
        if land_sea - sea_land == 1:
            continue
        S = _calc_decay_ps_value(track, p_landfall, land_sea - 1, s_rel)
        if S <= 1:
            # central_pressure at start of landfall > env_pres after landfall:
            # set central_pressure to environmental pressure during whole lf
            track.central_pressure[sea_land:land_sea] = track.environmental_pressure[sea_land:land_sea]
        else:
            p_decay = _decay_p_function(S, p_rel[ss_scale][1],
                                        track.dist_since_lf[sea_land:land_sea].values)
            # dont apply decay if it would decrease central pressure
            if np.any(p_decay < 1):
                LOGGER.info('Landfall decay would decrease pressure for '
                            'track id %s, leading to an intensification '
                            'of the Tropical Cyclone. This behaviour is '
                            'unphysical and therefore landfall decay is not '
                            'applied in this case.',
                            track.sid)
                p_decay[p_decay < 1] = track.central_pressure[sea_land:land_sea][p_decay < 1] / p_landfall
            track.central_pressure[sea_land:land_sea] = p_landfall * p_decay

        v_decay = _decay_v_function(v_rel[ss_scale],
                                    track.dist_since_lf[sea_land:land_sea].values)
        # dont apply decay if it would increase wind speeds
        if np.any(v_decay > 1):
            # should not happen unless v_rel is negative
            LOGGER.info('Landfall decay would increase wind speed for '
                        'track id %s. This behavious in unphysical and '
                        'therefore landfall decay is not applied in this '
                        'case.',
                        track.sid)
            v_decay[v_decay > 1] = (track.max_sustained_wind[sea_land:land_sea][v_decay > 1]
                                    / v_landfall)
        track.max_sustained_wind[sea_land:land_sea] = v_landfall * v_decay

        # correct values of sea after a landfall (until next landfall, if any)
        if land_sea < track.time.size:
            if idx + 1 < sea_land_idx.size:
                # if there is a next landfall, correct until last point before
                # reaching land again
                end_cor = sea_land_idx[idx + 1]
            else:
                # if there is no further landfall, correct until the end of
                # the track
                end_cor = track.time.size
            rndn = 0.1 * float(np.abs(np.random.normal(size=1) * 5) + 6)
            r_diff = track.central_pressure[land_sea].values - \
                     track.central_pressure[land_sea - 1].values + rndn
            track.central_pressure[land_sea:end_cor] += - r_diff

            rndn = rndn * 10  # mean value 10
            r_diff = track.max_sustained_wind[land_sea].values - \
                     track.max_sustained_wind[land_sea - 1].values - rndn
            track.max_sustained_wind[land_sea:end_cor] += - r_diff

        # correct limits
        np.warnings.filterwarnings('ignore')
        cor_p = track.central_pressure.values > track.environmental_pressure.values
        track.central_pressure[cor_p] = track.environmental_pressure[cor_p]
        track.max_sustained_wind[track.max_sustained_wind < 0] = 0

    track.attrs['category'] = climada.hazard.tc_tracks.set_category(
        track.max_sustained_wind.values, track.max_sustained_wind_unit)
    return track


def _check_apply_decay_plot(all_tracks, syn_orig_wind, syn_orig_pres):
    """Plot wind and presure before and after correction for synthetic tracks.
    Plot wind and presure for unchanged historical tracks."""
    # Plot synthetic tracks
    sy_tracks = [track for track in all_tracks if not track.orig_event_flag]
    graph_v_b, graph_v_a, graph_p_b, graph_p_a, graph_pd_a, graph_ped_a = \
        _check_apply_decay_syn_plot(sy_tracks, syn_orig_wind,
                                    syn_orig_pres)

    # Plot historic tracks
    hist_tracks = [track for track in all_tracks if track.orig_event_flag]
    graph_hv, graph_hp, graph_hpd_a, graph_hped_a = \
        _check_apply_decay_hist_plot(hist_tracks)

    # Put legend and fix size
    scale_thresholds = climada.hazard.tc_tracks.SAFFIR_SIM_CAT
    leg_lines = [Line2D([0], [0], color=climada.hazard.tc_tracks.CAT_COLORS[i_col], lw=2)
                 for i_col in range(len(scale_thresholds))]
    leg_lines.append(Line2D([0], [0], color='k', lw=2))
    leg_names = [climada.hazard.tc_tracks.CAT_NAMES[i_col]
                 for i_col in sorted(climada.hazard.tc_tracks.CAT_NAMES.keys())]
    leg_names.append('Sea')
    all_gr = [graph_v_a, graph_v_b, graph_p_a, graph_p_b, graph_ped_a,
              graph_pd_a, graph_hv, graph_hp, graph_hpd_a, graph_hped_a]
    for graph in all_gr:
        graph.axs[0].legend(leg_lines, leg_names)
        fig, _ = graph.get_elems()
        fig.set_size_inches(18.5, 10.5)


def _calc_decay_ps_value(track, p_landfall, pos, s_rel):
    if s_rel:
        p_land_s = track.environmental_pressure[pos].values
    else:
        p_land_s = track.central_pressure[pos].values
    return float(p_land_s / p_landfall)


def _decay_v_function(a_coef, x_val):
    """Decay function used for wind after landfall."""
    return np.exp(-a_coef * x_val)


def _solve_decay_v_function(v_y, x_val):
    """Solve decay function used for wind after landfall. Get A coefficient."""
    return -np.log(v_y) / x_val


def _decay_p_function(s_coef, b_coef, x_val):
    """Decay function used for pressure after landfall."""
    return s_coef - (s_coef - 1) * np.exp(-b_coef * x_val)


def _solve_decay_p_function(ps_y, p_y, x_val):
    """Solve decay function used for pressure after landfall.
    Get B coefficient."""
    return -np.log((ps_y - p_y) / (ps_y - 1.0)) / x_val


def _check_apply_decay_syn_plot(sy_tracks, syn_orig_wind,
                                syn_orig_pres):
    """Plot winds and pressures of synthetic tracks before and after
    correction."""
    _, graph_v_b = plt.subplots()
    graph_v_b.set_title('Wind before land decay correction')
    graph_v_b.set_xlabel('Node number')
    graph_v_b.set_ylabel('Max sustained wind (kn)')

    _, graph_v_a = plt.subplots()
    graph_v_a.set_title('Wind after land decay correction')
    graph_v_a.set_xlabel('Node number')
    graph_v_a.set_ylabel('Max sustained wind (kn)')

    _, graph_p_b = plt.subplots()
    graph_p_b.set_title('Pressure before land decay correctionn')
    graph_p_b.set_xlabel('Node number')
    graph_p_b.set_ylabel('Central pressure (mb)')

    _, graph_p_a = plt.subplots()
    graph_p_a.set_title('Pressure after land decay correctionn')
    graph_p_a.set_xlabel('Node number')
    graph_p_a.set_ylabel('Central pressure (mb)')

    _, graph_pd_a = plt.subplots()
    graph_pd_a.set_title('Relative pressure after land decay correction')
    graph_pd_a.set_xlabel('Distance from landfall (km)')
    graph_pd_a.set_ylabel('Central pressure relative to landfall')

    _, graph_ped_a = plt.subplots()
    graph_ped_a.set_title(
        'Environmental - central pressure after land decay correction')
    graph_ped_a.set_xlabel('Distance from landfall (km)')
    graph_ped_a.set_ylabel('Environmental pressure - Central pressure (mb)')

    for track, orig_wind, orig_pres in \
            zip(sy_tracks, syn_orig_wind, syn_orig_pres):
        sea_land_idx, land_sea_idx = climada.hazard.tc_tracks._get_landfall_idx(track)
        if sea_land_idx.size:
            for sea_land, land_sea in zip(sea_land_idx, land_sea_idx):
                v_lf = track.max_sustained_wind[sea_land - 1].values
                p_lf = track.central_pressure[sea_land - 1].values
                scale_thresholds = climada.hazard.tc_tracks.SAFFIR_SIM_CAT
                ss_scale_idx = np.where(v_lf < scale_thresholds)[0][0]
                on_land = np.arange(track.time.size)[sea_land:land_sea]

                graph_v_a.plot(on_land, track.max_sustained_wind[on_land],
                               'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_v_b.plot(on_land, orig_wind[on_land],
                               'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_p_a.plot(on_land, track.central_pressure[on_land],
                               'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_p_b.plot(on_land, orig_pres[on_land],
                               'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_pd_a.plot(track.dist_since_lf[on_land],
                                track.central_pressure[on_land] / p_lf,
                                'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_ped_a.plot(track.dist_since_lf[on_land],
                                 track.environmental_pressure[on_land] -
                                 track.central_pressure[on_land],
                                 'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])

            on_sea = np.arange(track.time.size)[~track.on_land]
            graph_v_a.plot(on_sea, track.max_sustained_wind[on_sea],
                           'o', c='k', markersize=5)
            graph_v_b.plot(on_sea, orig_wind[on_sea],
                           'o', c='k', markersize=5)
            graph_p_a.plot(on_sea, track.central_pressure[on_sea],
                           'o', c='k', markersize=5)
            graph_p_b.plot(on_sea, orig_pres[on_sea],
                           'o', c='k', markersize=5)

    return graph_v_b, graph_v_a, graph_p_b, graph_p_a, graph_pd_a, graph_ped_a


def _check_apply_decay_hist_plot(hist_tracks):
    """Plot winds and pressures of historical tracks."""
    _, graph_hv = plt.subplots()
    graph_hv.set_title('Historical wind')
    graph_hv.set_xlabel('Node number')
    graph_hv.set_ylabel('Max sustained wind (kn)')

    _, graph_hp = plt.subplots()
    graph_hp.set_title('Historical pressure')
    graph_hp.set_xlabel('Node number')
    graph_hp.set_ylabel('Central pressure (mb)')

    _, graph_hpd_a = plt.subplots()
    graph_hpd_a.set_title('Historical relative pressure')
    graph_hpd_a.set_xlabel('Distance from landfall (km)')
    graph_hpd_a.set_ylabel('Central pressure relative to landfall')

    _, graph_hped_a = plt.subplots()
    graph_hped_a.set_title('Historical environmental - central pressure')
    graph_hped_a.set_xlabel('Distance from landfall (km)')
    graph_hped_a.set_ylabel('Environmental pressure - Central pressure (mb)')

    for track in hist_tracks:
        sea_land_idx, land_sea_idx = climada.hazard.tc_tracks._get_landfall_idx(track)
        if sea_land_idx.size:
            for sea_land, land_sea in zip(sea_land_idx, land_sea_idx):
                scale_thresholds = climada.hazard.tc_tracks.SAFFIR_SIM_CAT
                ss_scale_idx = np.where(track.max_sustained_wind[sea_land - 1].values
                                 < scale_thresholds)[0][0]
                on_land = np.arange(track.time.size)[sea_land:land_sea]

                graph_hv.add_curve(on_land, track.max_sustained_wind[on_land],
                                   'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_hp.add_curve(on_land, track.central_pressure[on_land],
                                   'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_hpd_a.plot(track.dist_since_lf[on_land],
                                 track.central_pressure[on_land]
                                 / track.central_pressure[sea_land - 1].values,
                                 'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])
                graph_hped_a.plot(track.dist_since_lf[on_land],
                                  track.environmental_pressure[on_land] -
                                  track.central_pressure[on_land],
                                  'o', c=climada.hazard.tc_tracks.CAT_COLORS[ss_scale_idx])

            on_sea = np.arange(track.time.size)[~track.on_land]
            graph_hp.plot(on_sea, track.central_pressure[on_sea],
                          'o', c='k', markersize=5)
            graph_hv.plot(on_sea, track.max_sustained_wind[on_sea],
                          'o', c='k', markersize=5)

    return graph_hv, graph_hp, graph_hpd_a, graph_hped_a



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test MeasureSet and Measure classes.
"""
import unittest
import copy
from pathlib import Path

import numpy as np

from climada import CONFIG
from climada.hazard.base import Hazard
from climada.entity.entity_def import Entity
from climada.entity.exposures.base import Exposures, INDICATOR_IMPF
from climada.entity.impact_funcs.impact_func_set import ImpactFuncSet
from climada.entity.impact_funcs.base import ImpactFunc
from climada.entity.measures.measure_set import MeasureSet
from climada.entity.measures.base import Measure, IMPF_ID_FACT
from climada.util.constants import EXP_DEMO_H5, HAZ_DEMO_H5
import climada.util.coordinates as u_coord
import climada.hazard.test as hazard_test
import climada.entity.exposures.test as exposures_test

DATA_DIR = CONFIG.measures.test_data.dir()

HAZ_TEST_MAT = Path(hazard_test.__file__).parent / 'data' / 'atl_prob_no_name.mat'
ENT_TEST_MAT = Path(exposures_test.__file__).parent / 'data' / 'demo_today.mat'

class TestApply(unittest.TestCase):
    """Test implement measures functions."""
    def test_change_imp_func_pass(self):
        """Test _change_imp_func"""
        meas = MeasureSet.from_mat(ENT_TEST_MAT)
        act_1 = meas.get_measure(name='Mangroves')[0]

        imp_set = ImpactFuncSet()
        imp_tc = ImpactFunc()
        imp_tc.haz_type = 'XX'
        imp_tc.id = 1
        imp_tc.intensity = np.arange(10, 100, 10)
        imp_tc.intensity[0] = 0.
        imp_tc.intensity[-1] = 100.
        imp_tc.mdd = np.array([0.0, 0.0, 0.021857142857143, 0.035887500000000,
                               0.053977415307403, 0.103534246575342, 0.180414000000000,
                               0.410796000000000, 0.410796000000000])
        imp_tc.paa = np.array([0, 0.005000000000000, 0.042000000000000, 0.160000000000000,
                               0.398500000000000, 0.657000000000000, 1.000000000000000,
                               1.000000000000000, 1.000000000000000])
        imp_set.append(imp_tc)
        new_imp = act_1._change_imp_func(imp_set).get_func('XX')[0]

        self.assertTrue(np.array_equal(new_imp.intensity, np.array([4., 24., 34., 44.,
            54., 64., 74., 84., 104.])))
        self.assertTrue(np.array_equal(new_imp.mdd, np.array([0, 0, 0.021857142857143, 0.035887500000000,
            0.053977415307403, 0.103534246575342, 0.180414000000000, 0.410796000000000, 0.410796000000000])))
        self.assertTrue(np.array_equal(new_imp.paa, np.array([0, 0.005000000000000, 0.042000000000000,
            0.160000000000000, 0.398500000000000, 0.657000000000000, 1.000000000000000,
            1.000000000000000, 1.000000000000000])))
        self.assertFalse(id(new_imp) == id(imp_tc))

    def test_cutoff_hazard_pass(self):
        """Test _cutoff_hazard_damage"""
        meas = MeasureSet.from_mat(ENT_TEST_MAT)
        act_1 = meas.get_measure(name='Seawall')[0]

        haz = Hazard.from_mat(HAZ_TEST_MAT)
        exp = Exposures.from_mat(ENT_TEST_MAT)
        exp.gdf.rename(columns={'impf': 'impf_TC'}, inplace=True)
        exp.check()

        imp_set = ImpactFuncSet.from_mat(ENT_TEST_MAT)

        new_haz = act_1._cutoff_hazard_damage(exp, imp_set, haz)

        self.assertFalse(id(new_haz) == id(haz))

        pos_no_null = np.array([6249, 7697, 9134, 13500, 13199, 5944, 9052, 9050, 2429,
                                5139, 9053, 7102, 4096, 1070, 5948, 1076, 5947, 7432,
                                5949, 11694, 5484, 6246, 12147, 778, 3326, 7199, 12498,
                               11698, 6245, 5327, 4819, 8677, 5970, 7101, 779, 3894,
                                9051, 5976, 3329, 5978, 4282, 11697, 7193, 5351, 7310,
                                7478, 5489, 5526, 7194, 4283, 7191, 5328, 4812, 5528,
                                5527, 5488, 7475, 5529, 776, 5758, 4811, 6223, 7479,
                                7470, 5480, 5325, 7477, 7318, 7317, 11696, 7313, 13165,
                                6221])
        all_haz = np.arange(haz.intensity.shape[0])
        all_haz[pos_no_null] = -1
        pos_null = np.argwhere(all_haz > 0).reshape(-1)
        for i_ev in pos_null:
            self.assertEqual(new_haz.intensity[i_ev, :].max(), 0)


    def test_cutoff_hazard_region_pass(self):
        """Test _cutoff_hazard_damage in specific region"""
        meas = MeasureSet.from_mat(ENT_TEST_MAT)
        act_1 = meas.get_measure(name='Seawall')[0]
        act_1.exp_region_id = [1]

        haz = Hazard.from_mat(HAZ_TEST_MAT)
        exp = Exposures.from_mat(ENT_TEST_MAT)
        exp.gdf['region_id'] = np.zeros(exp.gdf.shape[0])
        exp.gdf.region_id.values[10:] = 1
        exp.check()

        imp_set = ImpactFuncSet.from_mat(ENT_TEST_MAT)

        new_haz = act_1._cutoff_hazard_damage(exp, imp_set, haz)

        self.assertFalse(id(new_haz) == id(haz))

        pos_no_null = np.array([6249, 7697, 9134, 13500, 13199, 5944, 9052, 9050, 2429,
                                5139, 9053, 7102, 4096, 1070, 5948, 1076, 5947, 7432,
                                5949, 11694, 5484, 6246, 12147, 778, 3326, 7199, 12498,
                               11698, 6245, 5327, 4819, 8677, 5970, 7101, 779, 3894,
                                9051, 5976, 3329, 5978, 4282, 11697, 7193, 5351, 7310,
                                7478, 5489, 5526, 7194, 4283, 7191, 5328, 4812, 5528,
                                5527, 5488, 7475, 5529, 776, 5758, 4811, 6223, 7479,
                                7470, 5480, 5325, 7477, 7318, 7317, 11696, 7313, 13165,
                                6221])
        all_haz = np.arange(haz.intensity.shape[0])
        all_haz[pos_no_null] = -1
        pos_null = np.argwhere(all_haz > 0).reshape(-1)
        centr_null = np.unique(exp.gdf.centr_[exp.gdf.region_id == 0])
        for i_ev in pos_null:
            self.assertEqual(new_haz.intensity[i_ev, centr_null].max(), 0)

    def test_change_exposures_impf_pass(self):
        """Test _change_exposures_impf"""
        meas = Measure()
        meas.imp_fun_map = '1to3'
        meas.haz_type = 'TC'

        imp_set = ImpactFuncSet()
        imp_tc = ImpactFunc()
        imp_tc.haz_type = 'TC'
        imp_tc.id = 1
        imp_tc.intensity = np.arange(10, 100, 10)
        imp_tc.mdd = np.arange(10, 100, 10)
        imp_tc.paa = np.arange(10, 100, 10)
        imp_set.append(imp_tc)

        imp_tc = ImpactFunc()
        imp_tc.haz_type = 'TC'
        imp_tc.id = 3
        imp_tc.intensity = np.arange(10, 100, 10)
        imp_tc.mdd = np.arange(10, 100, 10) * 2
        imp_tc.paa = np.arange(10, 100, 10) * 2

        exp = Exposures.from_hdf5(EXP_DEMO_H5)
        new_exp = meas._change_exposures_impf(exp)

        self.assertEqual(new_exp.ref_year, exp.ref_year)
        self.assertEqual(new_exp.value_unit, exp.value_unit)
        self.assertEqual(new_exp.tag.file_name, exp.tag.file_name)
        self.assertEqual(new_exp.tag.description, exp.tag.description)
        self.assertTrue(np.array_equal(new_exp.gdf.value.values, exp.gdf.value.values))
        self.assertTrue(np.array_equal(new_exp.gdf.latitude.values, exp.gdf.latitude.values))
        self.assertTrue(np.array_equal(new_exp.gdf.longitude.values, exp.gdf.longitude.values))
        self.assertTrue(np.array_equal(exp.gdf[INDICATOR_IMPF + 'TC'].values, np.ones(new_exp.gdf.shape[0])))
        self.assertTrue(np.array_equal(new_exp.gdf[INDICATOR_IMPF + 'TC'].values, np.ones(new_exp.gdf.shape[0]) * 3))

    def test_change_all_hazard_pass(self):
        """Test _change_all_hazard method"""
        meas = Measure()
        meas.hazard_set = HAZ_DEMO_H5

        ref_haz = Hazard.from_hdf5(HAZ_DEMO_H5)

        hazard = Hazard('TC')
        new_haz = meas._change_all_hazard(hazard)

        self.assertEqual(new_haz.tag.file_name, ref_haz.tag.file_name)
        self.assertEqual(new_haz.tag.haz_type, ref_haz.tag.haz_type)
        self.assertTrue(np.array_equal(new_haz.frequency, ref_haz.frequency))
        self.assertTrue(np.array_equal(new_haz.date, ref_haz.date))
        self.assertTrue(np.array_equal(new_haz.orig, ref_haz.orig))
        self.assertTrue(np.array_equal(new_haz.centroids.coord, ref_haz.centroids.coord))
        self.assertTrue(np.array_equal(new_haz.intensity.data, ref_haz.intensity.data))
        self.assertTrue(np.array_equal(new_haz.fraction.data, ref_haz.fraction.data))

    def test_change_all_exposures_pass(self):
        """Test _change_all_exposures method"""
        meas = Measure()
        meas.exposures_set = EXP_DEMO_H5

        ref_exp = Exposures.from_hdf5(EXP_DEMO_H5)

        exposures = Exposures()
        exposures.gdf['latitude'] = np.ones(10)
        exposures.gdf['longitude'] = np.ones(10)
        new_exp = meas._change_all_exposures(exposures)

        self.assertEqual(new_exp.ref_year, ref_exp.ref_year)
        self.assertEqual(new_exp.value_unit, ref_exp.value_unit)
        self.assertEqual(new_exp.tag.file_name, ref_exp.tag.file_name)
        self.assertEqual(new_exp.tag.description, ref_exp.tag.description)
        self.assertTrue(np.array_equal(new_exp.gdf.value.values, ref_exp.gdf.value.values))
        self.assertTrue(np.array_equal(new_exp.gdf.latitude.values, ref_exp.gdf.latitude.values))
        self.assertTrue(np.array_equal(new_exp.gdf.longitude.values, ref_exp.gdf.longitude.values))

    def test_not_filter_exposures_pass(self):
        """Test _filter_exposures method with []"""
        meas = Measure()
        meas.exp_region_id = []

        exp = Exposures()
        imp_set = ImpactFuncSet()
        haz = Hazard('TC')

        new_exp = Exposures()
        new_impfs = ImpactFuncSet()
        new_haz = Hazard('TC')

        res_exp, res_ifs, res_haz = meas._filter_exposures(exp, imp_set, haz,
                                                           new_exp, new_impfs, new_haz)

        self.assertTrue(res_exp is new_exp)
        self.assertTrue(res_ifs is new_impfs)
        self.assertTrue(res_haz is new_haz)

        self.assertTrue(res_exp is not exp)
        self.assertTrue(res_ifs is not imp_set)
        self.assertTrue(res_haz is not haz)

    def test_filter_exposures_pass(self):
        """Test _filter_exposures method with two values"""
        meas = Measure()
        meas.exp_region_id = [3, 4]
        meas.haz_type = 'TC'

        exp = Exposures.from_mat(ENT_TEST_MAT)
        exp.gdf.rename(columns={'impf_': 'impf_TC', 'centr_': 'centr_TC'}, inplace=True)
        exp.gdf['region_id'] = np.ones(exp.gdf.shape[0])
        exp.gdf.region_id.values[:exp.gdf.shape[0] // 2] = 3
        exp.gdf.region_id[0] = 4
        exp.check()

        imp_set = ImpactFuncSet.from_mat(ENT_TEST_MAT)

        haz = Hazard.from_mat(HAZ_TEST_MAT)
        exp.assign_centroids(haz)

        new_exp = copy.deepcopy(exp)
        new_exp.gdf['value'] *= 3
        new_exp.gdf['impf_TC'].values[:20] = 2
        new_exp.gdf['impf_TC'].values[20:40] = 3
        new_exp.gdf['impf_TC'].values[40:] = 1

        new_ifs = copy.deepcopy(imp_set)
        new_ifs.get_func('TC')[1].intensity += 1
        ref_ifs = copy.deepcopy(new_ifs)

        new_haz = copy.deepcopy(haz)
        new_haz.intensity *= 4

        res_exp, res_ifs, res_haz = meas._filter_exposures(exp, imp_set, haz,
            new_exp.copy(deep=True), new_ifs, new_haz)

        # unchanged meta data
        self.assertEqual(res_exp.ref_year, exp.ref_year)
        self.assertEqual(res_exp.value_unit, exp.value_unit)
        self.assertEqual(res_exp.tag.file_name, exp.tag.file_name)
        self.assertEqual(res_exp.tag.description, exp.tag.description)
        self.assertTrue(u_coord.equal_crs(res_exp.crs, exp.crs))
        self.assertTrue(u_coord.equal_crs(res_exp.gdf.crs, exp.gdf.crs))

        # regions (that is just input data, no need for testing, but it makes the changed and unchanged parts obious)
        self.assertTrue(np.array_equal(res_exp.gdf.region_id.values[0], 4))
        self.assertTrue(np.array_equal(res_exp.gdf.region_id.values[1:25], np.ones(24) * 3))
        self.assertTrue(np.array_equal(res_exp.gdf.region_id.values[25:], np.ones(25)))

        # changed exposures
        self.assertTrue(np.array_equal(res_exp.gdf.value.values[:25], new_exp.gdf.value.values[:25]))
        self.assertTrue(np.all(np.not_equal(res_exp.gdf.value.values[:25], exp.gdf.value.values[:25])))
        self.assertTrue(np.all(np.not_equal(res_exp.gdf.impf_TC.values[:25], new_exp.gdf.impf_TC.values[:25])))
        self.assertTrue(np.array_equal(res_exp.gdf.latitude.values[:25], new_exp.gdf.latitude.values[:25]))
        self.assertTrue(np.array_equal(res_exp.gdf.longitude.values[:25], new_exp.gdf.longitude.values[:25]))

        # unchanged exposures
        self.assertTrue(np.array_equal(res_exp.gdf.value.values[25:], exp.gdf.value.values[25:]))
        self.assertTrue(np.all(np.not_equal(res_exp.gdf.value.values[25:], new_exp.gdf.value.values[25:])))
        self.assertTrue(np.array_equal(res_exp.gdf.impf_TC.values[25:], exp.gdf.impf_TC.values[25:]))
        self.assertTrue(np.array_equal(res_exp.gdf.latitude.values[25:], exp.gdf.latitude.values[25:]))
        self.assertTrue(np.array_equal(res_exp.gdf.longitude.values[25:], exp.gdf.longitude.values[25:]))

        # unchanged impact functions
        self.assertEqual(list(res_ifs.get_func().keys()), [meas.haz_type])
        self.assertEqual(res_ifs.get_func()[meas.haz_type][1].id, imp_set.get_func()[meas.haz_type][1].id)
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][1].intensity,
                                       imp_set.get_func()[meas.haz_type][1].intensity))
        self.assertEqual(res_ifs.get_func()[meas.haz_type][3].id, imp_set.get_func()[meas.haz_type][3].id)
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][3].intensity,
                                       imp_set.get_func()[meas.haz_type][3].intensity))

        # changed impact functions
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][1 + IMPF_ID_FACT].intensity,
                        ref_ifs.get_func()[meas.haz_type][1].intensity))
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][1 + IMPF_ID_FACT].paa,
                        ref_ifs.get_func()[meas.haz_type][1].paa))
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][1 + IMPF_ID_FACT].mdd,
                        ref_ifs.get_func()[meas.haz_type][1].mdd))
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][3 + IMPF_ID_FACT].intensity,
                        ref_ifs.get_func()[meas.haz_type][3].intensity))
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][3 + IMPF_ID_FACT].paa,
                        ref_ifs.get_func()[meas.haz_type][3].paa))
        self.assertTrue(np.array_equal(res_ifs.get_func()[meas.haz_type][3 + IMPF_ID_FACT].mdd,
                        ref_ifs.get_func()[meas.haz_type][3].mdd))

        # unchanged hazard
        self.assertTrue(np.array_equal(res_haz.intensity[:, :36].toarray(),
                        haz.intensity[:, :36].toarray()))
        self.assertTrue(np.array_equal(res_haz.intensity[:, 37:46].toarray(),
                        haz.intensity[:, 37:46].toarray()))
        self.assertTrue(np.array_equal(res_haz.intensity[:, 47:].toarray(),
                        haz.intensity[:, 47:].toarray()))

        # changed hazard
        self.assertTrue(np.array_equal(res_haz.intensity[[36, 46]].toarray(),
                        new_haz.intensity[[36, 46]].toarray()))

    def test_apply_ref_pass(self):
        """Test apply method: apply all measures but insurance"""
        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        entity.check()

        new_exp, new_ifs, new_haz = entity.measures.get_measure('TC', 'Mangroves').apply(entity.exposures,
            entity.impact_funcs, hazard)

        self.assertTrue(new_exp is entity.exposures)
        self.assertTrue(new_haz is hazard)
        self.assertFalse(new_ifs is entity.impact_funcs)

        new_imp = new_ifs.get_func('TC')[0]
        self.assertTrue(np.array_equal(new_imp.intensity, np.array([4., 24., 34., 44.,
            54., 64., 74., 84., 104.])))
        self.assertTrue(np.allclose(new_imp.mdd, np.array([0, 0, 0.021857142857143, 0.035887500000000,
            0.053977415307403, 0.103534246575342, 0.180414000000000, 0.410796000000000, 0.410796000000000])))
        self.assertTrue(np.allclose(new_imp.paa, np.array([0, 0.005000000000000, 0.042000000000000,
            0.160000000000000, 0.398500000000000, 0.657000000000000, 1.000000000000000,
            1.000000000000000, 1.000000000000000])))

        new_imp = new_ifs.get_func('TC')[1]
        self.assertTrue(np.array_equal(new_imp.intensity, np.array([4., 24., 34., 44.,
            54., 64., 74., 84., 104.])))
        self.assertTrue(np.allclose(new_imp.mdd, np.array([0, 0, 0, 0.025000000000000,
            0.054054054054054, 0.104615384615385, 0.211764705882353, 0.400000000000000, 0.400000000000000])))
        self.assertTrue(np.allclose(new_imp.paa, np.array([0, 0.004000000000000, 0, 0.160000000000000,
            0.370000000000000, 0.650000000000000, 0.850000000000000, 1.000000000000000,
            1.000000000000000])))

    def test_calc_impact_pass(self):
        """Test calc_impact method: apply all measures but insurance"""

        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.exposures.gdf.rename(columns={'impf': 'impf_TC'}, inplace=True)
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        entity.measures.get_measure(name='Mangroves', haz_type='TC').haz_type = 'TC'
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        entity.check()

        imp, risk_transf = entity.measures.get_measure('TC', 'Mangroves').calc_impact(
                entity.exposures, entity.impact_funcs, hazard)

        self.assertAlmostEqual(imp.aai_agg, 4.850407096284983e+09)
        self.assertAlmostEqual(imp.at_event[0], 0)
        self.assertAlmostEqual(imp.at_event[12], 1.470194187501225e+07)
        self.assertAlmostEqual(imp.at_event[41], 4.7226357936631286e+08)
        self.assertAlmostEqual(imp.at_event[11890], 1.742110428135755e+07)
        self.assertTrue(np.array_equal(imp.coord_exp[:, 0], entity.exposures.gdf.latitude))
        self.assertTrue(np.array_equal(imp.coord_exp[:, 1], entity.exposures.gdf.longitude))
        self.assertAlmostEqual(imp.eai_exp[0], 1.15677655725858e+08)
        self.assertAlmostEqual(imp.eai_exp[-1], 7.528669956120645e+07)
        self.assertAlmostEqual(imp.tot_value, 6.570532945599105e+11)
        self.assertEqual(imp.unit, 'USD')
        self.assertEqual(imp.imp_mat.shape, (0, 0))
        self.assertTrue(np.array_equal(imp.event_id, hazard.event_id))
        self.assertTrue(np.array_equal(imp.date, hazard.date))
        self.assertEqual(imp.event_name, hazard.event_name)
        self.assertEqual(imp.tag['exp'].file_name, entity.exposures.tag.file_name)
        self.assertEqual(imp.tag['haz'].file_name, hazard.tag.file_name)
        self.assertEqual(imp.tag['impf_set'].file_name, entity.impact_funcs.tag.file_name)
        self.assertEqual(risk_transf.aai_agg, 0)


    def test_calc_impact_transf_pass(self):
        """Test calc_impact method: apply all measures and insurance"""

        hazard = Hazard.from_mat(HAZ_TEST_MAT)

        entity = Entity.from_mat(ENT_TEST_MAT)
        entity.exposures.gdf.rename(columns={'impf': 'impf_TC'}, inplace=True)
        entity.measures._data['TC'] = entity.measures._data.pop('XX')
        for meas in entity.measures.get_measure('TC'):
            meas.haz_type = 'TC'
        meas = entity.measures.get_measure(name='Beach nourishment', haz_type='TC')
        meas.haz_type = 'TC'
        meas.hazard_inten_imp = (1, 0)
        meas.mdd_impact = (1, 0)
        meas.paa_impact = (1, 0)
        meas.risk_transf_attach = 5.0e8
        meas.risk_transf_cover = 1.0e9
        entity.check()

        imp, risk_transf = entity.measures.get_measure(name='Beach nourishment', haz_type='TC').calc_impact(
                entity.exposures, entity.impact_funcs, hazard)

        self.assertAlmostEqual(imp.aai_agg, 6.280804242609713e+09)
        self.assertAlmostEqual(imp.at_event[0], 0)
        self.assertAlmostEqual(imp.at_event[12], 8.648764833437817e+07)
        self.assertAlmostEqual(imp.at_event[41], 500000000)
        self.assertAlmostEqual(imp.at_event[11890], 6.498096646836635e+07)
        self.assertTrue(np.array_equal(imp.coord_exp, np.array([])))
        self.assertTrue(np.array_equal(imp.eai_exp, np.array([])))
        self.assertAlmostEqual(imp.tot_value, 6.570532945599105e+11)
        self.assertEqual(imp.unit, 'USD')
        self.assertEqual(imp.imp_mat.shape, (0, 0))
        self.assertTrue(np.array_equal(imp.event_id, hazard.event_id))
        self.assertTrue(np.array_equal(imp.date, hazard.date))
        self.assertEqual(imp.event_name, hazard.event_name)
        self.assertEqual(imp.tag['exp'].file_name, entity.exposures.tag.file_name)
        self.assertEqual(imp.tag['haz'].file_name, hazard.tag.file_name)
        self.assertEqual(imp.tag['impf_set'].file_name, entity.impact_funcs.tag.file_name)
        self.assertEqual(risk_transf.aai_agg, 2.3139691495470852e+08)

# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestApply)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

climada init
"""
from shutil import copyfile
from pathlib import Path

from .util.config import CONFIG
from .util.constants import *


GSDP_DIR = SYSTEM_DIR.joinpath('GSDP')

REPO_DATA = {
    'data/system': [
        ISIMIP_GPWV3_NATID_150AS,
        GLB_CENTROIDS_MAT,
        ENT_TEMPLATE_XLS,
        HAZ_TEMPLATE_XLS,
        RIVER_FLOOD_REGIONS_CSV,
        NATEARTH_CENTROIDS[150],
        NATEARTH_CENTROIDS[360],
        SYSTEM_DIR.joinpath('WEALTH2GDP_factors_CRI_2016.csv'),
        SYSTEM_DIR.joinpath('GDP_TWN_IMF_WEO_data.csv'),
        SYSTEM_DIR.joinpath('FAOSTAT_data_country_codes.csv'),
        SYSTEM_DIR.joinpath('rcp_db.xls'),
        SYSTEM_DIR.joinpath('tc_impf_cal_v01_TDR1.0.csv'),
        SYSTEM_DIR.joinpath('tc_impf_cal_v01_EDR.csv'),
        SYSTEM_DIR.joinpath('tc_impf_cal_v01_RMSF.csv'),
    ],
    'data/system/GSDP': [
        GSDP_DIR.joinpath(f'{cc}_GSDP.xls')
        for cc in ['AUS', 'BRA', 'CAN', 'CHE', 'CHN', 'DEU', 'FRA', 'IDN', 'IND', 'JPN', 'MEX',
                   'TUR', 'USA', 'ZAF']
    ],
    'data/demo': [
        ENT_DEMO_TODAY,
        ENT_DEMO_FUTURE,
        EXP_DEMO_H5,
        HAZ_DEMO_FL,
        HAZ_DEMO_MAT,
        HAZ_DEMO_H5,
        TC_ANDREW_FL,
        DEMO_DIR.joinpath('demo_emdat_impact_data_2020.csv'),
    ] + WS_DEMO_NC
}


def test_installation():
    """Run this function to check whether climada is working properly.
    If the invoked tests pass and an OK is printed out, the installation was successfull.
    """
    from unittest import TestLoader, TextTestRunner
    suite = TestLoader().discover(start_dir='climada.engine.test',
                                           pattern='test_cost_benefit.py')
    suite.addTest(TestLoader().discover(start_dir='climada.engine.test',
                                                 pattern='test_impact.py'))
    TextTestRunner(verbosity=2).run(suite)


def setup_climada_data(reload=False):
    """This function is called when climada is imported.
    It creates a climada directory by default in the home directory.
    Other locations can be configured in the climada.conf file.
    The directory is filled with data files from the repository and is also the default target
    directory for files downloaded from climada.ethz.ch via the data api.

    Parameters
    ----------
    reload : bool, optional
        in case system or demo data have changed in the github repository, the local copies of
        these files can be updated by setting reload to True,
        by default False
    """
    for dirpath in [DEMO_DIR, SYSTEM_DIR, GSDP_DIR]:
        dirpath.mkdir(parents=True, exist_ok=True)

    for src_dir, path_list in REPO_DATA.items():
        for path in path_list:
            if not path.exists() or reload:
                src = Path(__file__).parent.parent.joinpath(src_dir, path.name)
                copyfile(src, path)

setup_climada_data()



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define CostBenefit class.
"""

__all__ = ['CostBenefit', 'risk_aai_agg', 'risk_rp_100', 'risk_rp_250']

import copy
import logging
import numpy as np
import matplotlib.colors as colors
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, FancyArrowPatch
from tabulate import tabulate

from climada.engine.impact import Impact

LOGGER = logging.getLogger(__name__)

DEF_PRESENT_YEAR = 2016
"""Default present reference year"""

DEF_FUTURE_YEAR = 2030
"""Default future reference year"""

NO_MEASURE = 'no measure'
"""Name of risk metrics when no measure is applied"""

def risk_aai_agg(impact):
    """Risk measurement as average annual impact aggregated.

    Parameters
    ----------
    impact : climada.engine.Impact
        an Impact instance

    Returns
    -------
    float
    """
    return impact.aai_agg

def risk_rp_100(impact):
    """Risk measurement as exceedance impact at 100 years return period.

    Parameters
    ----------
    impact : climada.engine.Impact
        an Impact instance

    Returns
    -------
    float
    """
    if impact.at_event.size > 0:
        efc = impact.calc_freq_curve([100])
        return efc.impact[0]
    return 0

def risk_rp_250(impact):
    """Risk measurement as exceedance impact at 250 years return period.

    Parameters
    ----------
    impact : climada.engine.Impact
        an Impact instance

    Returns
    -------
    float
    """
    if impact.at_event.size > 0:
        efc = impact.calc_freq_curve([250])
        return efc.impact[0]
    return 0

class CostBenefit():
    """Impact definition. Compute from an entity (exposures and impact
    functions) and hazard.

    Attributes
    ----------
    present_year : int
        present reference year
    future_year : int
        future year
    tot_climate_risk : float
        total climate risk without measures
    unit : str
        unit used for impact
    color_rgb : dict
        color code RGB for each measure.
    Key : str
        measure name ('no measure' used for case without measure),
    Value : np.array
    benefit : dict
        benefit of each measure. Key: measure name, Value: float benefit
    cost_ben_ratio : dict
        cost benefit ratio of each measure. Key: measure
        name, Value: float cost benefit ratio
    imp_meas_future : dict
        impact of each measure at future or default.
        Key: measure name ('no measure' used for case without measure),
        Value: dict with:
        'cost' (tuple): (cost measure, cost factor insurance),
        'risk' (float): risk measurement,
        'risk_transf' (float): annual expected risk transfer,
        'efc'  (ImpactFreqCurve): impact exceedance freq (optional)
        'impact' (Impact): impact instance
    imp_meas_present : dict
        impact of each measure at present.
        Key: measure name ('no measure' used for case without measure),
        Value: dict with:
        'cost' (tuple): (cost measure, cost factor insurance),
        'risk' (float): risk measurement,
        'risk_transf' (float): annual expected risk transfer,
        'efc'  (ImpactFreqCurve): impact exceedance freq (optional)
        'impact' (Impact): impact instance
    """

    def __init__(self):
        """Initilization"""
        self.present_year = DEF_PRESENT_YEAR
        self.future_year = DEF_FUTURE_YEAR

        self.tot_climate_risk = 0.0
        self.unit = 'USD'

        # dictionaries with key: measure name
        # value: measure color_rgb
        self.color_rgb = dict()
        # value: measure benefit
        self.benefit = dict()
        # value: measure cost benefit
        self.cost_ben_ratio = dict()
        # 'no measure' key for impact without measures
        # values: dictionary with 'cost': cost measure,
        #                         'risk': risk measurement,
        #                         'efc': ImpactFreqCurve
        #          (optionally)   'impact': Impact
        self.imp_meas_future = dict()
        self.imp_meas_present = dict()

    def calc(self, hazard, entity, haz_future=None, ent_future=None,
             future_year=None, risk_func=risk_aai_agg, imp_time_depen=None, save_imp=False):
        """Compute cost-benefit ratio for every measure provided current
        and, optionally, future conditions. Present and future measures need
        to have the same name. The measures costs need to be discounted by the user.
        If future entity provided, only the costs of the measures
        of the future and the discount rates of the present will be used.

        Parameters
        ----------
        hazard : climada.Hazard
        entity : climada.entity
        haz_future : climada.Hazard, optional
            hazard in the future (future year provided at ent_future)
        ent_future : Entity, optional
            entity in the future. Default is None
        future_year : int, optional
            future year to consider if no ent_future. Default is None
            provided. The benefits are added from the entity.exposures.ref_year until
            ent_future.exposures.ref_year, or until future_year if no ent_future given.
            Default: entity.exposures.ref_year+1
        risk_func : func optional
            function describing risk measure to use
            to compute the annual benefit from the Impact.
            Default: average annual impact (aggregated).
        imp_time_depen : float, optional
            parameter which represents time
            evolution of impact (super- or sublinear). If None: all years
            count the same when there is no future hazard nor entity and 1
            (linear annual change) when there is future hazard or entity.
            Default is None.
        save_imp : bool, optional)
        True if Impact of each measure is saved. Default is False.
        """
        # Present year given in entity. Future year in ent_future if provided.
        self.present_year = entity.exposures.ref_year
        self.unit = entity.exposures.value_unit

        # save measure colors
        for meas in entity.measures.get_measure(hazard.tag.haz_type):
            self.color_rgb[meas.name] = meas.color_rgb
        self.color_rgb[NO_MEASURE] = colors.to_rgb('deepskyblue')

        if future_year is None and ent_future is None:
            future_year = entity.exposures.ref_year

        if not haz_future and not ent_future:
            self.future_year = future_year
            self._calc_impact_measures(hazard, entity.exposures,
                                       entity.measures, entity.impact_funcs, 'future',
                                       risk_func, save_imp)
        else:
            if imp_time_depen is None:
                imp_time_depen = 1
            self._calc_impact_measures(hazard, entity.exposures,
                                       entity.measures, entity.impact_funcs, 'present',
                                       risk_func, save_imp)
            if haz_future and ent_future:
                self.future_year = ent_future.exposures.ref_year
                self._calc_impact_measures(haz_future, ent_future.exposures,
                                           ent_future.measures, ent_future.impact_funcs, 'future',
                                           risk_func, save_imp)
            elif haz_future:
                self.future_year = future_year
                self._calc_impact_measures(haz_future, entity.exposures,
                                           entity.measures, entity.impact_funcs, 'future',
                                           risk_func, save_imp)
            else:
                self.future_year = ent_future.exposures.ref_year
                self._calc_impact_measures(hazard, ent_future.exposures,
                                           ent_future.measures, ent_future.impact_funcs, 'future',
                                           risk_func, save_imp)

        self._calc_cost_benefit(entity.disc_rates, imp_time_depen)
        self._print_results()
        self._print_npv()

    def combine_measures(self, in_meas_names, new_name, new_color, disc_rates,
                         imp_time_depen=None, risk_func=risk_aai_agg):
        """Compute cost-benefit of the combination of measures previously
        computed by calc with save_imp=True. The benefits of the
        measures per event are added. To combine with risk transfer options use
        apply_risk_transfer.

        Parameters
        ----------
        in_meas_names : list(str)
        list with names of measures to combine
        new_name :  str
            name to give to the new resulting measure
            new_color (np.array): color code RGB for new measure, e.g.
            np.array([0.1, 0.1, 0.1])
        disc_rates : DiscRates
            discount rates instance
        imp_time_depen : float, optional
            parameter which represents time
            evolution of impact (super- or sublinear). If None: all years
            count the same when there is no future hazard nor entity and 1
            (linear annual change) when there is future hazard or entity.
            Default is None.
        risk_func : func, optional
            function describing risk measure given
            an Impact. Default: average annual impact (aggregated).

        Returns
        -------
        climada.CostBenefit
        """
        new_cb = CostBenefit()
        new_cb.present_year = self.present_year
        new_cb.future_year = self.future_year
        new_cb.tot_climate_risk = self.tot_climate_risk
        new_cb.unit = self.unit
        new_cb.color_rgb[NO_MEASURE] = self.color_rgb[NO_MEASURE]
        new_cb.color_rgb[new_name] = new_color
        new_cb.imp_meas_future[NO_MEASURE] = self.imp_meas_future[NO_MEASURE]

        # compute impacts for imp_meas_future and imp_meas_present
        self._combine_imp_meas(new_cb, in_meas_names, new_name, risk_func, when='future')
        if self.imp_meas_present:
            new_cb.imp_meas_present[NO_MEASURE] = self.imp_meas_present[NO_MEASURE]
            if imp_time_depen is None:
                imp_time_depen = 1
            self._combine_imp_meas(new_cb, in_meas_names, new_name, risk_func, when='present')

        # cost-benefit computation: fill measure's benefit and cost_ben_ratio
        time_dep = new_cb._time_dependency_array(imp_time_depen)
        new_cb._cost_ben_one(new_name, new_cb.imp_meas_future[new_name], disc_rates,
                             time_dep)
        new_cb._print_results()
        new_cb._print_npv()
        return new_cb

    def apply_risk_transfer(self, meas_name, attachment, cover, disc_rates,
                            cost_fix=0, cost_factor=1, imp_time_depen=None,
                            risk_func=risk_aai_agg):
        """Applies risk transfer to given measure computed before with saved
        impact and compares it to when no measure is applied. Appended to
        dictionaries of measures.

        Parameters
        ----------
        meas_name : str
            name of measure where to apply risk transfer
        attachment : float
            risk transfer values attachment (deductible)
        cover : float
            risk transfer cover
        cost_fix : float
            fixed cost of implemented innsurance, e.g. transaction costs
        cost_factor : float, optional
            factor to which to multiply the insurance layer
            to compute its cost. Default is 1
        imp_time_depen : float, optional
            parameter which represents time
            evolution of impact (super- or sublinear). If None: all years
            count the same when there is no future hazard nor entity and 1
            (linear annual change) when there is future hazard or entity.
            Default is None.
        risk_func : func, optional
            function describing risk measure given
            an Impact. Default: average annual impact (aggregated).
        """
        m_transf_name = 'risk transfer (' + meas_name + ')'
        self.color_rgb[m_transf_name] = np.maximum(np.minimum(self.color_rgb[meas_name] -
                                                              np.ones(3) * 0.2, 1), 0)

        _, layer_no = self.imp_meas_future[NO_MEASURE]['impact']. \
            calc_risk_transfer(attachment, cover)
        layer_no = risk_func(layer_no)

        imp, layer = self.imp_meas_future[meas_name]['impact']. \
            calc_risk_transfer(attachment, cover)
        self.imp_meas_future[m_transf_name] = dict()
        self.imp_meas_future[m_transf_name]['risk_transf'] = risk_func(layer)
        self.imp_meas_future[m_transf_name]['impact'] = imp
        self.imp_meas_future[m_transf_name]['risk'] = risk_func(imp)
        self.imp_meas_future[m_transf_name]['cost'] = (cost_fix, cost_factor)
        self.imp_meas_future[m_transf_name]['efc'] = imp.calc_freq_curve()

        if self.imp_meas_present:
            if imp_time_depen is None:
                imp_time_depen = 1
            time_dep = self._time_dependency_array(imp_time_depen)
            _, pres_layer_no = self.imp_meas_present[NO_MEASURE]['impact']. \
                calc_risk_transfer(attachment, cover)
            pres_layer_no = risk_func(pres_layer_no)
            layer_no = pres_layer_no + (layer_no - pres_layer_no) * time_dep

            imp, layer = self.imp_meas_present[meas_name]['impact']. \
                calc_risk_transfer(attachment, cover)
            self.imp_meas_present[m_transf_name] = dict()
            self.imp_meas_present[m_transf_name]['risk_transf'] = risk_func(layer)
            self.imp_meas_present[m_transf_name]['impact'] = imp
            self.imp_meas_present[m_transf_name]['risk'] = risk_func(imp)
            self.imp_meas_present[m_transf_name]['cost'] = (cost_fix, cost_factor)
            self.imp_meas_present[m_transf_name]['efc'] = imp.calc_freq_curve()
        else:
            time_dep = self._time_dependency_array(imp_time_depen)
            layer_no = time_dep * layer_no

        self._cost_ben_one(m_transf_name, self.imp_meas_future[m_transf_name],
                           disc_rates, time_dep, ini_state=meas_name)

        # compare layer no measure
        layer_no = disc_rates.net_present_value(self.present_year,
                                                self.future_year, layer_no)
        layer = ((self.cost_ben_ratio[m_transf_name] * self.benefit[m_transf_name] - cost_fix)
                 / cost_factor)
        self._print_results()
        self._print_risk_transfer(layer, layer_no, cost_fix, cost_factor)
        self._print_npv()

    def remove_measure(self, meas_name):
        """Remove computed values of given measure

        Parameters
        ----------
        meas_name : str
            name of measure to remove
        """
        del self.color_rgb[meas_name]
        del self.benefit[meas_name]
        del self.cost_ben_ratio[meas_name]
        del self.imp_meas_future[meas_name]
        if self.imp_meas_present:
            del self.imp_meas_present[meas_name]

    def plot_cost_benefit(self, cb_list=None, axis=None, **kwargs):
        """Plot cost-benefit graph. Call after calc().

        Parameters
        ----------
        cb_list : list(CostBenefit), optional
            if other CostBenefit
            provided, overlay them all. Used for uncertainty visualization.
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for Rectangle matplotlib, e.g. alpha=0.5
            (color is set by measures color attribute)

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if cb_list:
            if 'alpha' not in kwargs:
                kwargs['alpha'] = 0.5
            cb_uncer = [self]
            cb_uncer.extend(cb_list)
            axis = self._plot_list_cost_ben(cb_uncer, axis, **kwargs)
            return axis

        if 'alpha' not in kwargs:
            kwargs['alpha'] = 1.0
        axis = self._plot_list_cost_ben([self], axis, **kwargs)
        norm_fact, norm_name = _norm_values(self.tot_climate_risk + 0.01)

        text_pos = self.imp_meas_future[NO_MEASURE]['risk'] / norm_fact
        axis.scatter(text_pos, 0, c='r', zorder=200, clip_on=False)
        axis.text(text_pos, 0, '  AAI', horizontalalignment='center',
                  verticalalignment='bottom', rotation=90, fontsize=12, color='r')
        if abs(text_pos - self.tot_climate_risk / norm_fact) > 1:
            axis.scatter(self.tot_climate_risk / norm_fact, 0, c='r', zorder=200, clip_on=False)
            axis.text(self.tot_climate_risk / norm_fact, 0, '  Tot risk',
                      horizontalalignment='center', verticalalignment='bottom', rotation=90,
                      fontsize=12, color='r')

        axis.set_xlim(0, max(self.tot_climate_risk / norm_fact,
                             np.array(list(self.benefit.values())).sum() / norm_fact))
        axis.set_ylim(0, int(1 / np.nanmin(np.ma.masked_equal(np.array(list(
            self.cost_ben_ratio.values())), 0))) + 1)

        x_label = ('NPV averted damage over ' + str(self.future_year - self.present_year + 1)
                   + ' years (' + self.unit + ' ' + norm_name + ')')
        axis.set_xlabel(x_label)
        axis.set_ylabel('Benefit/Cost ratio')
        return axis

    def plot_event_view(self, return_per=(10, 25, 100), axis=None, **kwargs):
        """Plot averted damages for return periods. Call after calc().

        Parameters
        ----------
        return_per : list, optional
            years to visualize. Default 10, 25, 100
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for bar matplotlib function, e.g. alpha=0.5
            (color is set by measures color attribute)

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if not self.imp_meas_future:
            raise ValueError('Compute CostBenefit.calc() first')
        if not axis:
            _, axis = plt.subplots(1, 1)
        avert_rp = dict()
        for meas_name, meas_val in self.imp_meas_future.items():
            if meas_name == NO_MEASURE:
                continue
            interp_imp = np.interp(return_per, meas_val['efc'].return_per,
                                   meas_val['efc'].impact)
            # check if measure over no measure or combined with another measure
            try:
                ref_meas = meas_name[meas_name.index('(') + 1:meas_name.index(')')]
            except ValueError:
                ref_meas = NO_MEASURE
            ref_imp = np.interp(return_per,
                                self.imp_meas_future[ref_meas]['efc'].return_per,
                                self.imp_meas_future[ref_meas]['efc'].impact)
            avert_rp[meas_name] = ref_imp - interp_imp

        m_names = list(self.cost_ben_ratio.keys())
        sort_cb = np.argsort(np.array([self.cost_ben_ratio[name] for name in m_names]))
        names_sort = [m_names[i] for i in sort_cb]
        color_sort = [self.color_rgb[name] for name in names_sort]
        ref_imp = np.interp(return_per, self.imp_meas_future[NO_MEASURE]['efc'].return_per,
                            self.imp_meas_future[NO_MEASURE]['efc'].impact)
        for rp_i, _ in enumerate(return_per):
            val_i = [avert_rp[name][rp_i] for name in names_sort]
            cum_effect = np.cumsum(np.array([0] + val_i))
            for (eff, color) in zip(cum_effect[::-1][:-1], color_sort[::-1]):
                axis.bar(rp_i + 1, eff, color=color, **kwargs)
            axis.bar(rp_i + 1, ref_imp[rp_i], edgecolor='k', fc=(1, 0, 0, 0), zorder=100)
        axis.set_xlabel('Return Period (%s)' % str(self.future_year))
        axis.set_ylabel('Impact (' + self.unit + ')')
        axis.set_xticks(np.arange(len(return_per)) + 1)
        axis.set_xticklabels([str(per) for per in return_per])
        return axis

    @staticmethod
    def plot_waterfall(hazard, entity, haz_future, ent_future,
                       risk_func=risk_aai_agg, axis=None, **kwargs):
        """Plot waterfall graph at future with given risk metric. Can be called
        before and after calc().

        Parameters
        ----------
        hazard : climada.Hazard
        entity : climada.Entity
        haz_future : Hazard
            hazard in the future (future year provided at ent_future)
        ent_future : climada.Entity
            entity in the future
        risk_func : func, optional
            function describing risk measure given
            an Impact. Default: average annual impact (aggregated).
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for bar matplotlib function, e.g. alpha=0.5

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if ent_future.exposures.ref_year == entity.exposures.ref_year:
            raise ValueError('Same reference years for future and present entities.')
        present_year = entity.exposures.ref_year
        future_year = ent_future.exposures.ref_year

        imp = Impact()
        imp.calc(entity.exposures, entity.impact_funcs, hazard)
        curr_risk = risk_func(imp)

        imp = Impact()
        imp.calc(ent_future.exposures, ent_future.impact_funcs, haz_future)
        fut_risk = risk_func(imp)

        if not axis:
            _, axis = plt.subplots(1, 1)
        norm_fact, norm_name = _norm_values(curr_risk)

        # current situation
        LOGGER.info('Risk at {:d}: {:.3e}'.format(present_year, curr_risk))

        # changing future
        # socio-economic dev
        imp = Impact()
        imp.calc(ent_future.exposures, ent_future.impact_funcs, hazard)
        risk_dev = risk_func(imp)
        LOGGER.info('Risk with development at {:d}: {:.3e}'.format(future_year, risk_dev))

        # socioecon + cc
        LOGGER.info('Risk with development and climate change at {:d}: {:.3e}'.
                    format(future_year, fut_risk))

        axis.bar(1, curr_risk / norm_fact, **kwargs)
        axis.text(1, curr_risk / norm_fact, str(int(round(curr_risk / norm_fact))),
                  horizontalalignment='center', verticalalignment='bottom',
                  fontsize=12, color='k')
        axis.bar(2, height=(risk_dev - curr_risk) / norm_fact,
                 bottom=curr_risk / norm_fact, **kwargs)
        axis.text(2, curr_risk / norm_fact + (risk_dev - curr_risk) / norm_fact / 2,
                  str(int(round((risk_dev - curr_risk) / norm_fact))),
                  horizontalalignment='center', verticalalignment='center', fontsize=12, color='k')
        axis.bar(3, height=(fut_risk - risk_dev) / norm_fact,
                 bottom=risk_dev / norm_fact, **kwargs)
        axis.text(3, risk_dev / norm_fact + (fut_risk - risk_dev) / norm_fact / 2,
                  str(int(round((fut_risk - risk_dev) / norm_fact))),
                  horizontalalignment='center', verticalalignment='center', fontsize=12,
                  color='k')
        axis.bar(4, height=fut_risk / norm_fact, **kwargs)
        axis.text(4, fut_risk / norm_fact, str(int(round(fut_risk / norm_fact))),
                  horizontalalignment='center', verticalalignment='bottom',
                  fontsize=12, color='k')

        axis.set_xticks(np.arange(4) + 1)
        axis.set_xticklabels(['Risk ' + str(present_year),
                              'Economic \ndevelopment',
                              'Climate \nchange',
                              'Risk ' + str(future_year)])
        axis.set_ylabel('Impact (' + imp.unit + ' ' + norm_name + ')')
        axis.set_title('Risk at {:d} and {:d}'.format(present_year, future_year))
        return axis

    def plot_arrow_averted(self, axis, in_meas_names=None, accumulate=False, combine=False,
                           risk_func=risk_aai_agg, disc_rates=None, imp_time_depen=1, **kwargs):
        """Plot waterfall graph with accumulated values from present to future
        year. Call after calc() with save_imp=True.

        Parameters
        ----------
        axis : matplotlib.axes._subplots.AxesSubplot
            axis from plot_waterfall
            or plot_waterfall_accumulated where arrow will be added to last bar
        in_meas_names : list(str), optional
            list with names of measures to
            represented total averted damage. Default: all measures
        accumulate : bool, optional)
            accumulated averted damage (True) or averted
            damage in future (False). Default: False
        combine : bool, optional
            use combine_measures to compute total averted
            damage (True) or just add benefits (False). Default: False
        risk_func : func, optional
            function describing risk measure given
            an Impact used in combine_measures. Default: average annual impact (aggregated).
        disc_rates : DiscRates, optional
            discount rates used in combine_measures
        imp_time_depen : float, optional
            parameter which represent time
            evolution of impact used in combine_measures. Default: 1 (linear).
        kwargs : optional
            arguments for bar matplotlib function, e.g. alpha=0.5
        """
        if not in_meas_names:
            in_meas_names = list(self.benefit.keys())
        bars = [rect for rect in axis.get_children() if isinstance(rect, Rectangle)]

        if accumulate:
            tot_benefit = np.array([self.benefit[meas] for meas in in_meas_names]).sum()
            norm_fact = self.tot_climate_risk / bars[3].get_height()
        else:
            tot_benefit = np.array([risk_func(self.imp_meas_future[NO_MEASURE]['impact']) -
                                    risk_func(self.imp_meas_future[meas]['impact'])
                                    for meas in in_meas_names]).sum()
            norm_fact = (risk_func(self.imp_meas_future['no measure']['impact'])
                         / bars[3].get_height())
        if combine:
            try:
                LOGGER.info('Combining measures %s', in_meas_names)
                all_meas = self.combine_measures(in_meas_names, 'combine',
                                                 colors.to_rgba('black'), disc_rates,
                                                 imp_time_depen, risk_func)
            except KeyError:
                LOGGER.warning('Use calc() with save_imp=True to get a more accurate '
                               'approximation of total averted damage,')
            if accumulate:
                tot_benefit = all_meas.benefit['combine']
            else:
                tot_benefit = risk_func(all_meas.imp_meas_future[NO_MEASURE]['impact']) - \
                    risk_func(all_meas.imp_meas_future['combine']['impact'])

        self._plot_averted_arrow(axis, bars[3], tot_benefit, bars[3].get_height() * norm_fact,
                                 norm_fact, **kwargs)

    def plot_waterfall_accumulated(self, hazard, entity, ent_future,
                                   risk_func=risk_aai_agg, imp_time_depen=1,
                                   axis=None, **kwargs):
        """Plot waterfall graph with accumulated values from present to future
        year. Call after calc() with save_imp=True. Provide same inputs as in calc.

        Parameters
        ----------
        hazard : climada.Hazard
        entity : climada.Entity
        ent_future : climada.Entity
            entity in the future
        risk_func : func, optional
            function describing risk measure given an Impact.
            Default: average annual impact (aggregated).
        imp_time_depen : float, optional
            parameter which represent time
            evolution of impact used in combine_measures. Default: 1 (linear).
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for bar matplotlib function, e.g. alpha=0.5

        Returns
        -------
            matplotlib.axes._subplots.AxesSubplot
        """
        if not self.imp_meas_future or not self.imp_meas_present:
            raise ValueError('Compute CostBenefit.calc() first')
        if ent_future.exposures.ref_year == entity.exposures.ref_year:
            raise ValueError('Same reference years for future and present entities.')

        self.present_year = entity.exposures.ref_year
        self.future_year = ent_future.exposures.ref_year

        # current situation
        curr_risk = self.imp_meas_present[NO_MEASURE]['risk']
        time_dep = self._time_dependency_array()
        risk_curr = self._npv_unaverted_impact(curr_risk, entity.disc_rates,
                                               time_dep)
        LOGGER.info('Current total risk at {:d}: {:.3e}'.format(self.future_year,
                                                                risk_curr))

        # changing future
        time_dep = self._time_dependency_array(imp_time_depen)
        # socio-economic dev
        imp = Impact()
        imp.calc(ent_future.exposures, ent_future.impact_funcs, hazard)
        risk_dev = self._npv_unaverted_impact(risk_func(imp), entity.disc_rates,
                                              time_dep, curr_risk)
        LOGGER.info('Total risk with development at {:d}: {:.3e}'.format(
            self.future_year, risk_dev))

        # socioecon + cc
        risk_tot = self._npv_unaverted_impact(self.imp_meas_future[NO_MEASURE]['risk'],
                                              entity.disc_rates, time_dep, curr_risk)
        LOGGER.info('Total risk with development and climate change at {:d}: {:.3e}'.
                    format(self.future_year, risk_tot))

        # plot
        if not axis:
            _, axis = plt.subplots(1, 1)
        norm_fact, norm_name = _norm_values(curr_risk)
        axis.bar(1, risk_curr / norm_fact, **kwargs)
        axis.text(1, risk_curr / norm_fact, str(int(round(risk_curr / norm_fact))),
                  horizontalalignment='center', verticalalignment='bottom',
                  fontsize=12, color='k')
        axis.bar(2, height=(risk_dev - risk_curr) / norm_fact,
                 bottom=risk_curr / norm_fact, **kwargs)
        axis.text(2, risk_curr / norm_fact + (risk_dev - risk_curr) / norm_fact / 2,
                  str(int(round((risk_dev - risk_curr) / norm_fact))),
                  horizontalalignment='center', verticalalignment='center', fontsize=12, color='k')
        axis.bar(3, height=(risk_tot - risk_dev) / norm_fact,
                 bottom=risk_dev / norm_fact, **kwargs)
        axis.text(3, risk_dev / norm_fact + (risk_tot - risk_dev) / norm_fact / 2,
                  str(int(round((risk_tot - risk_dev) / norm_fact))),
                  horizontalalignment='center', verticalalignment='center', fontsize=12, color='k')
        axis.bar(4, height=risk_tot / norm_fact, **kwargs)
        axis.text(4, risk_tot / norm_fact, str(int(round(risk_tot / norm_fact))),
                  horizontalalignment='center', verticalalignment='bottom',
                  fontsize=12, color='k')

        axis.set_xticks(np.arange(4) + 1)
        axis.set_xticklabels(['Risk ' + str(self.present_year),
                              'Economic \ndevelopment',
                              'Climate \nchange',
                              'Risk ' + str(self.future_year)])
        axis.set_ylabel('Impact (' + self.unit + ' ' + norm_name + ')')
        axis.set_title('Total accumulated impact from {:d} to {:d}'.format(
            self.present_year, self.future_year))
        return axis

    def _calc_impact_measures(self, hazard, exposures, meas_set, imp_fun_set,
                              when='future', risk_func=risk_aai_agg, save_imp=False):
        """Compute impact of each measure and transform it to input risk
        measurement. Set reference year from exposures value.

        Parameters
        ----------
        hazard : climada.Hazard
        exposures : climada.entity.Exposures
        meas_set : climada.MeasureSet
            set of measures.
        imp_fun_set : ImpactFuncSet
            set of impact functions.
        when : str, optional
            'present' or 'future'. The conditions that
            are being considered.
        risk_func : func, optional
            function describing risk measure given an Impact.
            Default: average annual impact (aggregated).
        save_imp : bool, optional
            activate if Impact of each measure is
            saved. Default: False.
        """
        impact_meas = dict()

        # compute impact without measures
        LOGGER.debug('%s impact with no measure.', when)
        imp_tmp = Impact()
        imp_tmp.calc(exposures, imp_fun_set, hazard)
        impact_meas[NO_MEASURE] = dict()
        impact_meas[NO_MEASURE]['cost'] = (0, 0)
        impact_meas[NO_MEASURE]['risk'] = risk_func(imp_tmp)
        impact_meas[NO_MEASURE]['risk_transf'] = 0.0
        impact_meas[NO_MEASURE]['efc'] = imp_tmp.calc_freq_curve()
        if save_imp:
            impact_meas[NO_MEASURE]['impact'] = imp_tmp

        # compute impact for each measure
        for measure in meas_set.get_measure(hazard.tag.haz_type):
            LOGGER.debug('%s impact of measure %s.', when, measure.name)
            imp_tmp, risk_transf = measure.calc_impact(exposures, imp_fun_set, hazard)
            impact_meas[measure.name] = dict()
            impact_meas[measure.name]['cost'] = (measure.cost, measure.risk_transf_cost_factor)
            impact_meas[measure.name]['risk'] = risk_func(imp_tmp)
            impact_meas[measure.name]['risk_transf'] = risk_func(risk_transf)
            impact_meas[measure.name]['efc'] = imp_tmp.calc_freq_curve()
            if save_imp:
                impact_meas[measure.name]['impact'] = imp_tmp

        # if present reference provided save it
        if when == 'future':
            self.imp_meas_future = impact_meas
        else:
            self.imp_meas_present = impact_meas

    def _calc_cost_benefit(self, disc_rates, imp_time_depen=None):
        """Compute discounted impact from present year to future year

        Parameters
        ----------
        disc_rates : DiscRates
            discount rates instance
        imp_time_depen : float, optional
            parameter which represent time evolution of impact
        """
        LOGGER.info('Computing cost benefit from years %s to %s.',
                    str(self.present_year), str(self.future_year))

        if self.future_year - self.present_year + 1 <= 0:
            raise ValueError('Wrong year range: %s - %s.'
                             % (str(self.present_year), str(self.future_year)))

        if not self.imp_meas_future:
            raise ValueError('Compute first _calc_impact_measures')

        time_dep = self._time_dependency_array(imp_time_depen)

        # discounted cost benefit for each measure and total climate risk
        for meas_name, meas_val in self.imp_meas_future.items():
            if meas_name == NO_MEASURE:
                # npv of the full unaverted damages
                if self.imp_meas_present:
                    self.tot_climate_risk = self._npv_unaverted_impact(
                        self.imp_meas_future[NO_MEASURE]['risk'],
                        disc_rates, time_dep, self.imp_meas_present[NO_MEASURE]['risk'])
                else:
                    self.tot_climate_risk = self._npv_unaverted_impact(
                        self.imp_meas_future[NO_MEASURE]['risk'],
                        disc_rates, time_dep)
                continue

            self._cost_ben_one(meas_name, meas_val, disc_rates, time_dep)

    def _cost_ben_one(self, meas_name, meas_val, disc_rates, time_dep,
                      ini_state=NO_MEASURE):
        """Compute cost and benefit for given measure with time dependency

        Parameters
        ----------
        meas_name : str
            name of measure
        meas_val : dict
            contains measure's cost, risk, efc, risk_trans and
            optionally impact at future
        disc_rates : DiscRates
            discount rates instance
        time_dep : np.array
            time dependency array
        ini_state : str, optional
            name of the measure to which to compute benefit.
            Default: 'no measure'
        """
        fut_benefit = self.imp_meas_future[ini_state]['risk'] - meas_val['risk']
        fut_risk_tr = meas_val['risk_transf']
        if self.imp_meas_present:
            pres_benefit = self.imp_meas_present[ini_state]['risk'] - \
                self.imp_meas_present[meas_name]['risk']
            meas_ben = pres_benefit + (fut_benefit - pres_benefit) * time_dep

            pres_risk_tr = self.imp_meas_present[meas_name]['risk_transf']
            risk_tr = pres_risk_tr + (fut_risk_tr - pres_risk_tr) * time_dep
        else:
            meas_ben = time_dep * fut_benefit
            risk_tr = time_dep * fut_risk_tr

        # discount
        meas_ben = disc_rates.net_present_value(self.present_year,
                                                self.future_year, meas_ben)
        risk_tr = disc_rates.net_present_value(self.present_year,
                                               self.future_year, risk_tr)
        self.benefit[meas_name] = meas_ben
        with np.errstate(divide='ignore'):
            self.cost_ben_ratio[meas_name] = (meas_val['cost'][0]
                                              + meas_val['cost'][1] * risk_tr) / meas_ben

    def _time_dependency_array(self, imp_time_depen=None):
        """Construct time dependency array. Each year contains a value in [0,1]
        representing the rate of damage difference achieved that year, according
        to the growth represented by parameter imp_time_depen.

        Parameters
        ----------
         imp_time_depen : float, optional
            parameter which represent time evolution of impact

        Returns
        -------
        np.array
        """
        n_years = self.future_year - self.present_year + 1
        if imp_time_depen:
            time_dep = np.arange(n_years)**imp_time_depen / \
                (n_years - 1)**imp_time_depen
        else:
            time_dep = np.ones(n_years)
        return time_dep

    def _npv_unaverted_impact(self, risk_future, disc_rates, time_dep,
                              risk_present=None):
        """Net present value of total unaverted damages

        Parameters
        ----------
        risk_future : float
            risk under future situation
        disc_rates : climada.DiscRates
            discount rates object
        time_dep : np.array
            values in 0-1 indicating impact growth at each year
        risk_present : float
            risk under current situation

        Returns
        -------
        float
        """
        if risk_present:
            tot_climate_risk = risk_present + (risk_future - risk_present) * time_dep
            tot_climate_risk = disc_rates.net_present_value(self.present_year,
                                                            self.future_year,
                                                            tot_climate_risk)
        else:
            tot_climate_risk = disc_rates.net_present_value(self.present_year,
                                                            self.future_year,
                                                            time_dep * risk_future)
        return tot_climate_risk

    def _combine_imp_meas(self, new_cb, in_meas_names, new_name, risk_func, when='future'):
        """Compute impacts combined measures assuming they are independent, i.e.
        their benefit can be added. Costs are also added. For the new measure
        the dictionary imp_meas_future if when='future' and imp_meas_present
        if when='present'.

        Parameters
        ----------
        in_meas_names : list(str)
            list with names of measures to combine
        new_name : str
            name to give to the new resulting measure
        risk_func : func, optional
            function describing risk measure given
            an Impact. Default: average annual impact (aggregated).
        when : str, optional
            'present' or 'future' making reference to which dictionary
            to fill (imp_meas_present or imp_meas_future respectively)
            default: 'future'
        """
        if when == 'future':
            imp_dict = self.imp_meas_future
            new_imp_dict = new_cb.imp_meas_future
        else:
            imp_dict = self.imp_meas_present
            new_imp_dict = new_cb.imp_meas_present

        sum_ben = np.sum([
            imp_dict[NO_MEASURE]['impact'].at_event - imp_dict[name]['impact'].at_event
            for name in in_meas_names
        ], axis=0)
        new_imp = copy.deepcopy(imp_dict[in_meas_names[0]]['impact'])
        new_imp.at_event = np.maximum(imp_dict[NO_MEASURE]['impact'].at_event
                                      - sum_ben, 0)

        new_imp.eai_exp = np.array([])
        new_imp.aai_agg = sum(new_imp.at_event * new_imp.frequency)

        new_imp_dict[new_name] = dict()
        new_imp_dict[new_name]['impact'] = new_imp
        new_imp_dict[new_name]['efc'] = new_imp.calc_freq_curve()
        new_imp_dict[new_name]['risk'] = risk_func(new_imp)
        new_imp_dict[new_name]['cost'] = (
            np.array([imp_dict[name]['cost'][0] for name in in_meas_names]).sum(),
            1)
        new_imp_dict[new_name]['risk_transf'] = 0

    def _print_results(self):
        """Print table with main results"""
        norm_fact, norm_name = _norm_values(np.array(list(self.benefit.values())).max())
        norm_name = '(' + self.unit + ' ' + norm_name + ')'

        table = []
        headers = ['Measure', 'Cost ' + norm_name, 'Benefit ' + norm_name, 'Benefit/Cost']
        for meas_name in self.benefit:
            if not np.isnan(self.cost_ben_ratio[meas_name]) and \
            not np.isinf(self.cost_ben_ratio[meas_name]):
                cost = self.cost_ben_ratio[meas_name] * self.benefit[meas_name] / norm_fact
            else:
                cost = self.imp_meas_future[meas_name]['cost'][0] / norm_fact
            table.append([meas_name, cost, self.benefit[meas_name] / norm_fact,
                          1 / self.cost_ben_ratio[meas_name]])
        print()
        print(tabulate(table, headers, tablefmt="simple"))

        table = []
        table.append(['Total climate risk:',
                      self.tot_climate_risk / norm_fact, norm_name])
        table.append(['Average annual risk:',
                      self.imp_meas_future[NO_MEASURE]['risk'] / norm_fact, norm_name])
        table.append(['Residual risk:',
                      (self.tot_climate_risk -
                       np.array(list(self.benefit.values())).sum()) / norm_fact, norm_name])
        print()
        print(tabulate(table, tablefmt="simple"))

    @staticmethod
    def _plot_list_cost_ben(cb_list, axis=None, **kwargs):
        """Overlay cost-benefit bars for every measure

        Parameters
        ----------
        cb_list : list
            list of CostBenefit instances with filled values
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        kwargs : optional
            arguments for Rectangle matplotlib, e.g. alpha=0.5
            (color is set by measures color attribute)

        Returns
        -------
        matplotlib.axes._subplots.AxesSubplot
        """
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.5
        norm_fact = [_norm_values(cb_res.tot_climate_risk)[0] for cb_res in cb_list]
        norm_fact = np.array(norm_fact).mean()
        _, norm_name = _norm_values(norm_fact + 0.01)

        if not axis:
            _, axis = plt.subplots(1, 1)
        m_names = list(cb_list[0].cost_ben_ratio.keys())
        sort_cb = np.argsort(np.array([cb_list[0].cost_ben_ratio[name] for name in m_names]))
        xy_lim = [0, 0]
        for i_cb, cb_res in enumerate(cb_list):
            xmin = 0
            for meas_id in sort_cb:
                meas_n = m_names[meas_id]
                axis.add_patch(Rectangle((xmin, 0),
                                         cb_res.benefit[meas_n] / norm_fact,
                                         1 / cb_res.cost_ben_ratio[meas_n],
                                         color=cb_res.color_rgb[meas_n], **kwargs))

                if i_cb == 0:
                    axis.text(xmin + (cb_res.benefit[meas_n] / norm_fact) / 2,
                              0, '  ' + meas_n, horizontalalignment='center',
                              verticalalignment='bottom', rotation=90, fontsize=12)
                xmin += cb_res.benefit[meas_n] / norm_fact

            xy_lim[0] = max(xy_lim[0],
                            max(int(cb_res.tot_climate_risk / norm_fact),
                                np.array(list(cb_res.benefit.values())).sum() / norm_fact))
            try:
                with np.errstate(divide='ignore'):
                    xy_lim[1] = max(xy_lim[1], int(1 / cb_res.cost_ben_ratio[
                        m_names[sort_cb[0]]]) + 1)
            except (ValueError, OverflowError):
                xy_lim[1] = max(xy_lim[1],
                                int(1 / np.array(list(cb_res.cost_ben_ratio.values())).max()) + 1)

        axis.set_xlim(0, xy_lim[0])
        axis.set_ylim(0, xy_lim[1])
        axis.set_xlabel('NPV averted damage over ' +
                        str(cb_list[0].future_year - cb_list[0].present_year + 1) +
                        ' years (' + cb_list[0].unit + ' ' + norm_name + ')')
        axis.set_ylabel('Benefit/Cost ratio')
        return axis

    @staticmethod
    def _plot_averted_arrow(axis, bar_4, tot_benefit, risk_tot, norm_fact, **kwargs):
        """Plot arrow inn fourth bar of total averted damage by implementing
        all the measures.

        Parameters
        ----------
        axis : matplotlib.axes._subplots.AxesSubplot, optional
            axis to use
        bar_4 : matplotlib.container.BarContainer
            bar where arrow is plotted
        tot_benefit : float
            arrow length
        risk_tot : float
            total risk
        norm_fact : float
            normalization factor
        kwargs : optional
            arguments for bar matplotlib function, e.g. alpha=0.5
        """
        bar_bottom, bar_top = bar_4.get_bbox().get_points()
        axis.text(bar_top[0] - (bar_top[0] - bar_bottom[0]) / 2, bar_top[1],
                  "Averted", ha="center", va="top", rotation=270, size=15)
        arrow_len = min(tot_benefit / norm_fact, risk_tot / norm_fact)

        if 'color' not in kwargs:
            kwargs['color'] = 'k'
        if 'alpha' not in kwargs:
            kwargs['alpha'] = 0.4
        if 'mutation_scale' not in kwargs:
            kwargs['mutation_scale'] = 100
        axis.add_patch(FancyArrowPatch(
            (bar_top[0] - (bar_top[0] - bar_bottom[0]) / 2, bar_top[1]),
            (bar_top[0] - (bar_top[0] - bar_bottom[0]) / 2, risk_tot / norm_fact - arrow_len),
            **kwargs))

    def _print_risk_transfer(self, layer, layer_no, cost_fix, cost_factor):
        """Print comparative of risk transfer with and without measure

        Parameters
        ----------
        layer : float
            expected insurance layer with measure
        layer_on : float
            expected insurance layer without measure
        """
        norm_fact, norm_name = _norm_values(np.array(list(self.benefit.values())).max())
        norm_name = '(' + self.unit + ' ' + norm_name + ')'
        headers = ['Risk transfer', 'Expected damage in \n insurance layer ' +
                   norm_name, 'Price ' + norm_name]
        table = [['without measure', layer_no / norm_fact,
                  (cost_fix + layer_no * cost_factor) / norm_fact],
                 ['with measure', layer / norm_fact,
                  (cost_fix + layer * cost_factor) / norm_fact]]
        print()
        print(tabulate(table, headers, tablefmt="simple"))
        print()

    @staticmethod
    def _print_npv():
        print('Net Present Values')

def _norm_values(value):
    """Compute normalization value and name

    Parameters
    ----------
    value : float
        value to normalize

    Returns :
    norm_fact: float
    norm_name: float
    """
    norm_fact = 1.
    norm_name = ''
    if value / 1.0e9 > 1:
        norm_fact = 1.0e9
        norm_name = 'bn'
    elif value / 1.0e6 > 1:
        norm_fact = 1.0e6
        norm_name = 'm'
    elif value / 1.0e3 > 1:
        norm_fact = 1.0e3
        norm_name = 'k'
    return norm_fact, norm_name



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Define climate change scenarios for tropical cycones.
"""

import numpy as np
import pandas as pd

from climada.util.constants import SYSTEM_DIR

TOT_RADIATIVE_FORCE = SYSTEM_DIR.joinpath('rcp_db.xls')
"""© RCP Database (Version 2.0.5) http://www.iiasa.ac.at/web-apps/tnt/RcpDb.
generated: 2018-07-04 10:47:59."""

def get_knutson_criterion():
    """
    Fill changes in TCs according to Knutson et al. 2015 Global projections
    of intense tropical cyclone activity for the late twenty-first century from
    dynamical downscaling of CMIP5/RCP4.5 scenarios.

    Returns
    -------
    criterion : list(dict)
        list of the criterion dictionary for frequency and intensity change
        per basin, per category taken from the Table 3 in Knutson et al. 2015.
        with items 'basin' (str), 'category' (list(int)), 'year' (int),
        'change' (float), 'variable' ('intensity' or 'frequency')
    """
    # NA
    na = [
        {'basin': 'NA', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NA', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NA', 'category': [3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NA', 'category': [4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NA', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.045, 'variable': 'intensity'}
        ]

    # EP
    ep = [
        {'basin': 'EP', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.163, 'variable': 'frequency'},
        {'basin': 'EP', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.193, 'variable': 'frequency'},
        {'basin': 'EP', 'category': [3, 4, 5],
         'year': 2100, 'change': 1.837, 'variable': 'frequency'},
        {'basin': 'EP', 'category': [4, 5],
         'year': 2100, 'change': 3.375, 'variable': 'frequency'},
        {'basin': 'EP', 'category': [0],
         'year': 2100, 'change': 1.082, 'variable': 'intensity'},
        {'basin': 'EP', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.078, 'variable': 'intensity'}
        ]

    # WP
    wp = [
        {'basin': 'WP', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.345, 'variable': 'frequency'},
        {'basin': 'WP', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.316, 'variable': 'frequency'},
        {'basin': 'WP', 'category': [3, 4, 5],
         'year': 2100, 'change': 1 - 0.169, 'variable': 'frequency'},
        {'basin': 'WP', 'category': [4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'WP', 'category': [0],
         'year': 2100, 'change': 1.074, 'variable': 'intensity'},
        {'basin': 'WP', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.055, 'variable': 'intensity'},
        ]

    # SP
    sp = [
        {'basin': 'SP', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.366, 'variable': 'frequency'},
        {'basin': 'SP', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.406, 'variable': 'frequency'},
        {'basin': 'SP', 'category': [3, 4, 5],
         'year': 2100, 'change': 1 - 0.506, 'variable': 'frequency'},
        {'basin': 'SP', 'category': [4, 5],
         'year': 2100, 'change': 1 - 0.583, 'variable': 'frequency'}
        ]

    # NI
    ni = [
        {'basin': 'NI', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NI', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.256, 'variable': 'frequency'},
        {'basin': 'NI', 'category': [3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'NI', 'category': [4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'}
        ]

    # SI
    si = [
        {'basin': 'SI', 'category': [0, 1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.261, 'variable': 'frequency'},
        {'basin': 'SI', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1 - 0.284, 'variable': 'frequency'},
        {'basin': 'SI', 'category': [3, 4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'SI', 'category': [4, 5],
         'year': 2100, 'change': 1, 'variable': 'frequency'},
        {'basin': 'SI', 'category': [1, 2, 3, 4, 5],
         'year': 2100, 'change': 1.033, 'variable': 'intensity'}
        ]

    return na + ep + wp + sp + ni + si


def calc_scale_knutson(ref_year=2050, rcp_scenario=45):
    """
    Comparison 2081-2100 (i.e., late twenty-first century) and 2001-20
    (i.e., present day). Late twenty-first century effects on intensity and
    frequency per Saffir-Simpson-category and ocean basin is scaled to target
    year and target RCP proportional to total radiative forcing of the
    respective RCP and year.

    Parameters
    ----------
    ref_year : int, optional
        year between 2000 ad 2100. Default: 2050
    rcp_scenario: int, optional
        26 for RCP 2.6, 45 for RCP 4.5. The default is 45
        60 for RCP 6.0 and 85 for RCP 8.5.

    Returns
    -------
    factor : float
        factor to scale Knutson parameters to the give RCP and year
    """
    # Parameters used in Knutson et al 2015
    base_knu = np.arange(2001, 2021)
    end_knu = np.arange(2081, 2101)
    rcp_knu = 45

    # radiative forcings for each RCP scenario
    rad_force = pd.read_excel(TOT_RADIATIVE_FORCE)
    years = np.array([year for year in rad_force.columns if isinstance(year, int)])
    rad_rcp = np.array([int(float(sce[sce.index('.') - 1:sce.index('.') + 2]) * 10)
                        for sce in rad_force.Scenario if isinstance(sce, str)])

    # mean values for Knutson values
    rf_vals = np.argwhere(rad_rcp == rcp_knu).reshape(-1)[0]
    rf_vals = np.array([rad_force.iloc[rf_vals][year] for year in years])
    rf_base = np.nanmean(np.interp(base_knu, years, rf_vals))
    rf_end = np.nanmean(np.interp(end_knu, years, rf_vals))

    # scale factor for ref_year and rcp_scenario
    rf_vals = np.argwhere(rad_rcp == rcp_scenario).reshape(-1)[0]
    rf_vals = np.array([rad_force.iloc[rf_vals][year] for year in years])
    rf_sel = np.interp(ref_year, years, rf_vals)
    return max((rf_sel - rf_base) / (rf_end - rf_base), 0)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Functionalities to handle HDF5 files. Used for MATLAB files as well.
"""

__all__ = ['read',
           'get_string',
           'get_str_from_ref',
           'get_list_str_from_ref',
           'get_sparse_csr_mat'
          ]

import logging
from scipy import sparse
import numpy as np
import h5py

def read(file_name, with_refs=False):
    """Load a hdf5 data structure from a file.

        Parameters
        ----------
        file_name :
            file to load
        with_refs :
            enable loading of the references. Default is unset,
            since it increments the execution time considerably.

        Returns
        -------
        contents :
            dictionary structure containing all the variables.

        Examples
        --------
        >>> # Contents contains the Matlab data in a dictionary.
        >>> contents = read("/pathto/dummy.mat")
        >>> # Contents contains the Matlab data and its reference in a dictionary.
        >>> contents = read("/pathto/dummy.mat", True)

        Raises
        ------
        Exception while reading
        """
    def get_group(group):
        """Recursive function to get variables from a group."""
        contents = {}
        for name, obj in list(group.items()):
            if isinstance(obj, h5py.Dataset):
                contents[name] = np.array(obj)
            elif isinstance(obj, h5py.Group):
                # it is a group, so call self recursively
                if with_refs or name != "#refs#":
                    contents[name] = get_group(obj)
            # other objects such as links are ignored
        return contents

    try:
        file = h5py.File(file_name, 'r')
        contents = get_group(file)
        file.close()
        return contents
    except OSError as err:
        raise OSError(f'Invalid file {file_name}: ' + str(err)) from err

def get_string(array):
    """Form string from input array of unisgned integers.

        Parameters
        ----------
        array :
            array of integers

        Returns
        -------
        string
    """
    return u''.join(chr(int(c)) for c in array)

def get_str_from_ref(file_name, var):
    """Form string from a reference HDF5 variable of the given file.

        Parameters
        ----------
        file_name :
            matlab file name
        var :
            HDF5 reference variable

        Returns
        -------
        string
    """
    file = h5py.File(file_name, 'r')
    obj = file[var]
    return get_string(obj)

def get_list_str_from_ref(file_name, var):
    """Form list of strings from a reference HDF5 variable of the given file.

        Parameters
        ----------
        file_name :
            matlab file name
        var :
            array of HDF5 reference variable

        Returns
        -------
        string
    """
    name_list = []
    file = h5py.File(file_name, 'r')
    for name in var:
        name_list.append(get_string(file[name[0]][:]).strip())
    return name_list

def get_sparse_csr_mat(mat_dict, shape):
    """Form sparse matrix from input hdf5 sparse matrix data type.

        Parameters
        ----------
        mat_dict :
            dictionary containing the sparse matrix information.
        shape :
            tuple describing output matrix shape.

        Returns
        -------
        sparse csr matrix
    """
    # Check if input has all the necessary data of a sparse matrix
    if ('data' not in mat_dict) or ('ir' not in mat_dict) or \
    ('jc' not in mat_dict):
        raise ValueError('Input data is not a sparse matrix.')

    return sparse.csc_matrix((mat_dict['data'], mat_dict['ir'],
                              mat_dict['jc']), shape).tocsr()

def to_string(str_or_bytes):
    """converts a bytes object into a string if necessary

    Parameters
    ----------
    str_or_bytes : str or bytes

    Returns
    -------
    str
        the original string if executed with a str object otherwise decoded bytes
    """
    # TODO: remove this method from the module and replace its use with the asstr() method
    # of hdf5 datasets, as soon as the h5py version is high enough for that.
    return str_or_bytes.decode() if isinstance(str_or_bytes, bytes) else str_or_bytes



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

Test CentroidsVector and CentroidsRaster classes.
"""
import unittest
from pathlib import Path

import numpy as np
import pandas as pd
import geopandas as gpd

from climada import CONFIG
from climada.hazard.centroids.centr import Centroids
from climada.util.constants import GLB_CENTROIDS_MAT, HAZ_TEMPLATE_XLS
import climada.hazard.test as hazard_test

HAZ_TEST_MAT = Path(hazard_test.__file__).parent / 'data' / 'atl_prob_no_name.mat'


class TestCentroidsReader(unittest.TestCase):
    """Test read functions Centroids"""

    def test_mat_pass(self):
        """Read a centroid mat file correctly."""
        centroids = Centroids.from_mat(HAZ_TEST_MAT)

        n_centroids = 100
        self.assertEqual(centroids.coord.shape, (n_centroids, 2))
        self.assertEqual(centroids.coord[0][0], 21)
        self.assertEqual(centroids.coord[0][1], -84)
        self.assertEqual(centroids.coord[n_centroids - 1][0], 30)
        self.assertEqual(centroids.coord[n_centroids - 1][1], -75)

    def test_mat_global_pass(self):
        """Test read GLB_CENTROIDS_MAT"""
        centroids = Centroids.from_mat(GLB_CENTROIDS_MAT)

        self.assertEqual(centroids.region_id[1062443], 35)
        self.assertEqual(centroids.region_id[170825], 28)

    def test_centroid_pass(self):
        """Read a centroid excel file correctly."""
        centroids = Centroids.from_excel(HAZ_TEMPLATE_XLS)

        n_centroids = 45
        self.assertEqual(centroids.coord.shape[0], n_centroids)
        self.assertEqual(centroids.coord.shape[1], 2)
        self.assertEqual(centroids.coord[0][0], -25.95)
        self.assertEqual(centroids.coord[0][1], 32.57)
        self.assertEqual(centroids.coord[n_centroids - 1][0], -24.7)
        self.assertEqual(centroids.coord[n_centroids - 1][1], 33.88)

    def test_base_grid(self):
        """Read new centroids using from_base_grid, then select by extent."""
        centroids = Centroids.from_base_grid(land=True, res_as=150)
        self.assertEqual(centroids.lat.size, 8858035)
        self.assertTrue(np.all(np.diff(centroids.lat) <= 0))

        count_sandwich = np.sum(centroids.region_id == 239)
        self.assertEqual(count_sandwich, 321)

        count_sgi = centroids.select(
            reg_id=239,
            extent=(-39, -34.7, -55.5, -53.6)  # south georgia island
        ).size
        self.assertEqual(count_sgi, 296)

        # test negative latitudinal orientation by testing that northern hemisphere (Russia)
        # is listed before southern hemisphere (South Africa)
        russia_max_idx = (centroids.region_id == 643).nonzero()[0].max()
        safrica_min_idx = (centroids.region_id == 710).nonzero()[0].min()
        self.assertTrue(russia_max_idx < safrica_min_idx)

    def test_geodataframe(self):
        """Test that constructing a valid Centroids instance from gdf works."""
        gdf = gpd.GeoDataFrame(pd.read_excel(HAZ_TEMPLATE_XLS))
        gdf.geometry = gpd.points_from_xy(
                gdf['longitude'], gdf['latitude']
        )
        gdf['elevation'] = np.random.rand(gdf.geometry.size)
        gdf['region_id'] = np.zeros(gdf.geometry.size)
        gdf['region_id'][0] = np.NaN
        gdf['geom'] = gdf.geometry  # this should have no effect on centroids

        centroids = Centroids.from_geodataframe(gdf)
        centroids.check()

        self.assertEqual(centroids.geometry.size, 45)
        self.assertEqual(centroids.lon[0], 32.57)
        self.assertEqual(centroids.lat[0], -25.95)
        self.assertEqual(centroids.elevation.size, 45)
        self.assertEqual(centroids.on_land.sum(), 44)
        self.assertIsInstance(centroids.geometry, gpd.GeoSeries)
        self.assertIsInstance(centroids.geometry.total_bounds, np.ndarray)

class TestCentroidsMethods(unittest.TestCase):

    def test_union(self):
        cent1 = Centroids()
        cent1.lat, cent1.lon = np.array([0, 1]), np.array([0, -1])
        cent1.on_land = np.array([True, True])

        cent2 = Centroids()
        cent2.lat, cent2.lon = np.array([2, 3]), np.array([-2, 3])
        cent2.on_land = np.array([False, False])

        cent3 = Centroids()
        cent3.lat, cent3.lon = np.array([-1, -2]), np.array([1, 2])

        cent = cent1.union(cent2)
        np.testing.assert_array_equal(cent.lat, [0, 1, 2, 3])
        np.testing.assert_array_equal(cent.lon, [0, -1, -2, 3])
        np.testing.assert_array_equal(cent.on_land, [True, True, False, False])

        cent = cent1.union(cent1, cent2)
        np.testing.assert_array_equal(cent.lat, [0, 1, 2, 3])
        np.testing.assert_array_equal(cent.lon, [0, -1, -2, 3])
        np.testing.assert_array_equal(cent.on_land, [True, True, False, False])

        cent = Centroids.union(cent1)
        np.testing.assert_array_equal(cent.lat, [0, 1])
        np.testing.assert_array_equal(cent.lon, [0, -1])
        np.testing.assert_array_equal(cent.on_land, [True, True])

        cent = cent1.union(cent1)
        np.testing.assert_array_equal(cent.lat, [0, 1])
        np.testing.assert_array_equal(cent.lon, [0, -1])
        np.testing.assert_array_equal(cent.on_land, [True, True])

        cent = Centroids.union(cent1, cent2, cent3)
        np.testing.assert_array_equal(cent.lat, [0, 1, 2, 3, -1, -2])
        np.testing.assert_array_equal(cent.lon, [0, -1, -2, 3, 1, 2])


    def test_union_meta(self):
        cent1 = Centroids.from_pnt_bounds((-1, -1, 0, 0), res=1)
        cent2 = Centroids.from_pnt_bounds((0, 0, 1, 1), res=1)
        cent3 = Centroids.from_lat_lon(np.array([1]), np.array([1]))

        cent = cent1.union(cent2)
        np.testing.assert_array_equal(cent.lat, [0,  0, -1, -1,  1,  1,  0])
        np.testing.assert_array_equal(cent.lon, [-1,  0, -1,  0,  0,  1,  1])

        cent = cent3.union(cent1)
        np.testing.assert_array_equal(cent.lat, [1,  0,  0, -1, -1])
        np.testing.assert_array_equal(cent.lon, [1, -1,  0, -1,  0])


# Execute Tests
if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestCentroidsReader)
    TESTS.addTests(unittest.TestLoader().loadTestsFromTestCase(TestCentroidsMethods))
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of CLIMADA.

Copyright (C) 2017 ETH Zurich, CLIMADA contributors listed in AUTHORS.

CLIMADA is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.

CLIMADA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with CLIMADA. If not, see <https://www.gnu.org/licenses/>.

---

module containing functions to support various select methods.
"""



import logging
import numpy as np

LOGGER = logging.getLogger(__name__)

def get_attributes_with_matching_dimension(obj, dims):
    """
    Get the attributes of an object that have len(dims) number
    of dimensions or more, and all dims are individual parts of the
    attribute's shape.


    Parameters
    ----------
    obj : object of any class
        The object from which matching attributes are returned
    dims : list[int]
        List of dimensions size to match

    Returns
    -------
    list_of_attrs : list[str]
        List of names of the attributes with matching dimensions
    """

    list_of_attrs = []
    for attr, value in obj.__dict__.items():

        try:

            if isinstance(value, list):
                value = np.array(value)

            if all([dims.count(i) <= value.shape.count(i) for i in set(dims)]):
                list_of_attrs.append(attr)

        except AttributeError:
            pass

    return list_of_attrs
